#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from abc import abstractmethod
from typing import Any

# Bokeh imports
from ...core.has_props import abstract
from ...core.properties import (
    Dict,
    Either,
    Float,
    List,
    Nullable,
    Override,
    Required,
    String,
    Tuple,
)
from ...model import Model

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "Angular",
    "ImperialLength",
    "Metric",
    "MetricLength",
    "ReciprocalMetric",
    "ReciprocalMetricLength",
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Dimensional(Model):
    """ A base class for models defining units of measurement.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    ticks = Required(List(Float), help="""
    Preferred values to choose from in non-exact mode.
    """)

    include = Nullable(List(String), default=None, help="""
    An optional subset of preferred units from the basis.
    """)

    exclude = List(String, default=[], help="""
    A subset of units from the basis to avoid.
    """)

    @abstractmethod
    def is_known(self, unit: str) -> bool:
        pass

class CustomDimensional(Dimensional):
    """ A base class for units of measurement with an explicit basis.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    basis = Required(Dict(String, Either(Tuple(Float, String), Tuple(Float, String, String))), help="""
    The basis defining the units of measurement.

    This consists of a mapping between short unit names and their corresponding
    scaling factors, TeX names and optional long names. For example, the basis
    for defining angular units of measurement is:

    .. code-block:: python

        basis = {
            "°":  (1,      "^\\circ",           "degree"),
            "'":  (1/60,   "^\\prime",          "minute"),
            "''": (1/3600, "^{\\prime\\prime}", "second"),
        }
    """)

    def is_known(self, unit: str) -> bool:
        return unit in self.basis

class Metric(Dimensional):
    """ Model for defining metric units of measurement.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    base_unit = Required(String, help="""
    The short name of the base unit, e.g. ``"m"`` for meters or ``"eV"`` for electron volts.
    """)

    full_unit = Nullable(String, default=None, help="""
    The full name of the base unit, e.g. ``"meter"`` or ``"electronvolt"``.
    """)

    ticks = Override(default=[1, 2, 5, 10, 15, 20, 25, 50, 75, 100, 125, 150, 200, 250, 500, 750])

    def is_known(self, unit: str) -> bool:
        prefixes = ["Q", "R", "Y", "Z", "E", "P", "T", "G", "M", "k", "h", "", "d", "c", "m", "µ", "n", "p", "f", "a", "z", "y", "r", "q"]
        basis = {f"{prefix}{unit}" for prefix in prefixes}
        return unit in basis

class ReciprocalMetric(Metric):
    """ Model for defining reciprocal metric units of measurement, e.g. ``m^{-1}``.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

class MetricLength(Metric):
    """ Metric units of length measurement.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    base_unit = Override(default="m")

    exclude = Override(default=["dm", "hm"])

class ReciprocalMetricLength(ReciprocalMetric):
    """ Metric units of reciprocal length measurement.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    base_unit = Override(default="m")

    exclude = Override(default=["dm", "hm"])

class ImperialLength(CustomDimensional):
    """ Imperial units of length measurement.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    basis = Override(default={
        "in":  ( 1/12, "in",  "inch"   ),
        "ft":  (    1, "ft",  "foot"   ),
        "yd":  (    3, "yd",  "yard"   ),
        "ch":  (   66, "ch",  "chain"  ),
        "fur": (  660, "fur", "furlong"),
        "mi":  ( 5280, "mi",  "mile"   ),
        "lea": (15840, "lea", "league" ),
    })

    ticks = Override(default=[1, 3, 6, 12, 60])

class Angular(CustomDimensional):
    """ Units of angular measurement.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    basis = Override(default={
        "°":  (1,      "^\\circ",           "degree"),
        "'":  (1/60,   "^\\prime",          "minute"),
        "''": (1/3600, "^{\\prime\\prime}", "second"),
    })

    ticks = Override(default=[1, 3, 6, 12, 60, 120, 240, 360])

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import hashlib
import json
import os
from dataclasses import dataclass
from os.path import normpath
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterator,
    Sequence,
    TypedDict,
)
from urllib.parse import urljoin

# Bokeh imports
from ..core.has_props import HasProps
from ..core.templates import CSS_RESOURCES, JS_RESOURCES
from ..document.document import Document
from ..resources import Resources
from ..settings import settings
from ..util.compiler import bundle_models
from .util import contains_tex_string

if TYPE_CHECKING:
    from typing_extensions import NotRequired

    from ..resources import Hashes

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'Bundle',
    'bundle_for_objs_and_resources',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

class Artifact:
    pass

class ScriptRef(Artifact):
    def __init__(self, url: str, type: str = "text/javascript") -> None:
        self.url = URL(url)
        self.type = type


class Script(Artifact):
    def __init__(self, content: str, type: str = "text/javascript") -> None:
        self.content = content
        self.type = type


class StyleRef(Artifact):
    def __init__(self, url: str) -> None:
        self.url = URL(url)


class Style(Artifact):
    def __init__(self, content: str) -> None:
        self.content = content


class Bundle:

    js_files: list[URL]
    js_raw: list[str]
    css_files: list[URL]
    css_raw: list[str]
    hashes: Hashes

    def __init__(self, js_files: list[URL] = [], js_raw: list[str] = [],
            css_files: list[URL] = [], css_raw: list[str] = [], hashes: Hashes = {}):
        self.js_files = js_files[:]
        self.js_raw = js_raw[:]
        self.css_files = css_files[:]
        self.css_raw = css_raw[:]
        self.hashes = {**hashes}

    def __iter__(self) -> Iterator[str]:
        yield self._render_js()
        yield self._render_css()

    def _render_js(self) -> str:
        return JS_RESOURCES.render(js_files=self.js_files, js_raw=self.js_raw, hashes=self.hashes)

    def _render_css(self) -> str:
        return CSS_RESOURCES.render(css_files=self.css_files, css_raw=self.css_raw)

    def scripts(self, tag: bool = True) -> str:
        if tag:
            return JS_RESOURCES.render(js_raw=self.js_raw, js_files=[])
        else:
            return "\n".join(self.js_raw)

    @property
    def js_urls(self) -> list[URL]:
        return self.js_files

    @property
    def css_urls(self) -> list[URL]:
        return self.css_files

    def add(self, artifact: Artifact) -> None:
        if isinstance(artifact, ScriptRef):
            self.js_files.append(artifact.url)
        elif isinstance(artifact, Script):
            self.js_raw.append(artifact.content)
        elif isinstance(artifact, StyleRef):
            self.css_files.append(artifact.url)
        elif isinstance(artifact, Style):
            self.css_raw.append(artifact.content)

def bundle_for_objs_and_resources(objs: Sequence[HasProps | Document] | None, resources: Resources | None) -> Bundle:
    ''' Generate rendered CSS and JS resources suitable for the given
    collection of Bokeh objects

    Args:
        objs (seq[HasProps or Document]) :

        resources (Resources)

    Returns:
        Bundle

    '''
    if objs is not None:
        all_objs    = _all_objs(objs)
        use_widgets = _use_widgets(all_objs)
        use_tables  = _use_tables(all_objs)
        use_gl      = _use_gl(all_objs)
        use_mathjax = _use_mathjax(all_objs)
    else:
        # XXX: force all components on server and in notebook, because we don't know in advance what will be used
        all_objs    = None
        use_widgets = True
        use_tables  = True
        use_gl      = True
        use_mathjax = True

    js_files: list[URL] = []
    js_raw: list[str] = []
    css_files: list[URL] = []
    css_raw: list[str] = []

    if resources is not None:
        components = list(resources.components)
        if not use_widgets: components.remove("bokeh-widgets")
        if not use_tables:  components.remove("bokeh-tables")
        if not use_gl:      components.remove("bokeh-gl")
        if not use_mathjax: components.remove("bokeh-mathjax")

        resources = resources.clone(components=components)

        js_files.extend(map(URL, resources.js_files))
        js_raw.extend(resources.js_raw)

        css_files.extend(map(URL, resources.css_files))
        css_raw.extend(resources.css_raw)

        extensions = _bundle_extensions(all_objs if objs else None, resources)
        mode = resources.mode
        if mode == "inline":
            js_raw.extend([ Resources._inline(bundle.artifact_path) for bundle in extensions ])
        elif mode == "server":
            js_files.extend([ bundle.server_url for bundle in extensions ])
        elif mode == "cdn":
            for bundle in extensions:
                if bundle.cdn_url is not None:
                    js_files.append(bundle.cdn_url)
                else:
                    js_raw.append(Resources._inline(bundle.artifact_path))
        else:
            js_files.extend([ URL(str(bundle.artifact_path)) for bundle in extensions ])

    models = [ obj.__class__ for obj in all_objs ] if all_objs else None
    ext = bundle_models(models)
    if ext is not None:
        js_raw.append(ext)

    return Bundle(js_files, js_raw, css_files, css_raw, resources.hashes if resources else {})

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

def _query_extensions(all_objs: set[HasProps], query: Callable[[type[HasProps]], bool]) -> bool:
    names: set[str] = set()

    for obj in all_objs:
        if hasattr(obj, "__implementation__"):
            continue
        name = obj.__view_module__.split(".")[0]
        if name == "bokeh":
            continue
        if name in names:
            continue
        names.add(name)

        for model in HasProps.model_class_reverse_map.values():
            if model.__module__.startswith(name):
                if query(model):
                    return True

    return False

@dataclass(frozen=True)
class URL:
    """ Opaque type for representing URLs. """

    url: str

    def __truediv__(self, path: str) -> URL:
        url = self.url if self.url.endswith("/") else f"{self.url}/"
        return URL(urljoin(url, path.replace(os.sep, "/")))

    def __str__(self) -> str:
        return self.url

@dataclass(frozen=True)
class ExtensionEmbed:
    artifact_path: Path
    server_url: URL
    cdn_url: URL | None = None

class Pkg(TypedDict):
    name: NotRequired[str]
    version: NotRequired[str]
    module: NotRequired[str]
    main: NotRequired[str]

_default_cdn_host = URL("https://unpkg.com")

extension_dirs: dict[str, Path] = {}

def _bundle_extensions(objs: set[HasProps] | None, resources: Resources) -> list[ExtensionEmbed]:
    names: set[str] = set()
    bundles: list[ExtensionEmbed] = []

    extensions = [".min.js", ".js"] if resources.minified else [".js"]

    all_objs = objs if objs is not None else HasProps.model_class_reverse_map.values()

    for obj in all_objs:
        if hasattr(obj, "__implementation__"):
            continue
        name = obj.__view_module__.split(".")[0]
        if name == "bokeh":
            continue
        if name in names:
            continue
        names.add(name)
        module = __import__(name)
        this_file = Path(module.__file__).absolute()
        base_dir = this_file.parent
        dist_dir = base_dir / "dist"

        ext_path = base_dir / "bokeh.ext.json"
        if not ext_path.exists():
            continue

        server_prefix = URL(resources.root_url) / "static" / "extensions"
        package_path = base_dir / "package.json"

        pkg: Pkg | None = None
        if package_path.exists():
            with open(package_path) as io:
                try:
                    pkg = json.load(io)
                except json.decoder.JSONDecodeError:
                    pass

        artifact_path: Path
        server_url: URL
        cdn_url: URL | None = None

        if pkg is not None:
            pkg_name: str | None = pkg.get("name", None)
            if pkg_name is None:
                raise ValueError("invalid package.json; missing package name")
            pkg_version = pkg.get("version", "latest")
            pkg_main = pkg.get("module", pkg.get("main", None))
            if pkg_main is not None:
                pkg_main = Path(normpath(pkg_main))
                cdn_url = _default_cdn_host / f"{pkg_name}@{pkg_version}" / f"{pkg_main}"
            else:
                pkg_main = dist_dir / f"{name}.js"
            artifact_path = base_dir / pkg_main
            artifacts_dir = artifact_path.parent
            artifact_name = artifact_path.name
            server_path = f"{name}/{artifact_name}"
            if not settings.dev:
                sha = hashlib.sha256()
                sha.update(pkg_version.encode())
                vstring = sha.hexdigest()
                server_path = f"{server_path}?v={vstring}"
        else:
            for ext in extensions:
                artifact_path = dist_dir / f"{name}{ext}"
                artifacts_dir = dist_dir
                server_path = f"{name}/{name}{ext}"
                if artifact_path.exists():
                    break
            else:
                raise ValueError(f"can't resolve artifact path for '{name}' extension")

        extension_dirs[name] = Path(artifacts_dir)
        server_url = server_prefix / server_path
        embed = ExtensionEmbed(artifact_path, server_url, cdn_url)
        bundles.append(embed)

    return bundles

def _all_objs(objs: Sequence[HasProps | Document]) -> set[HasProps]:
    all_objs: set[HasProps] = set()

    for obj in objs:
        if isinstance(obj, Document):
            for root in obj.roots:
                all_objs |= root.references()
        else:
            all_objs |= obj.references()

    return all_objs

def _any(objs: set[HasProps], query: Callable[[HasProps], bool]) -> bool:
    ''' Whether any of a collection of objects satisfies a given query predicate

    Args:
        objs (set[HasProps]) :

        query (callable)

    Returns:
        True, if ``query(obj)`` is True for some object in ``objs``, else False

    '''
    return any(query(x) for x in objs)

def _use_tables(all_objs: set[HasProps]) -> bool:
    ''' Whether a collection of Bokeh objects contains a TableWidget

    Args:
        objs (seq[HasProps or Document]) :

    Returns:
        bool

    '''
    from ..models.widgets import TableWidget
    return _any(all_objs, lambda obj: isinstance(obj, TableWidget)) or _ext_use_tables(all_objs)

def _use_widgets(all_objs: set[HasProps]) -> bool:
    ''' Whether a collection of Bokeh objects contains a any Widget

    Args:
        objs (seq[HasProps or Document]) :

    Returns:
        bool

    '''
    from ..models.widgets import Widget
    return _any(all_objs, lambda obj: isinstance(obj, Widget)) or _ext_use_widgets(all_objs)

def _model_requires_mathjax(model: HasProps) -> bool:
    """Whether a model requires MathJax to be loaded
    Args:
        model (HasProps): HasProps to check
    Returns:
        bool: True if MathJax required, False if not
    """
    # TODO query model's properties that include TextLike or better
    # yet load mathjax bundle dynamically on bokehjs' side.

    from ..models.annotations import TextAnnotation
    from ..models.axes import Axis
    from ..models.widgets.markups import Div, Paragraph
    from ..models.widgets.sliders import AbstractSlider

    if isinstance(model, TextAnnotation):
        if isinstance(model.text, str) and contains_tex_string(model.text):
            return True

    if isinstance(model, AbstractSlider):
        if isinstance(model.title, str) and contains_tex_string(model.title):
            return True

    if isinstance(model, Axis):
        if isinstance(model.axis_label, str) and contains_tex_string(model.axis_label):
            return True

        for val in model.major_label_overrides.values():
            if isinstance(val, str) and contains_tex_string(val):
                return True

    if isinstance(model, Div) and not model.disable_math and not model.render_as_text:
        if contains_tex_string(model.text):
            return True

    if isinstance(model, Paragraph) and not model.disable_math:
        if contains_tex_string(model.text):
            return True

    return False

def _use_mathjax(all_objs: set[HasProps]) -> bool:
    ''' Whether a collection of Bokeh objects contains a model requesting MathJax
    Args:
        objs (seq[HasProps or Document]) :
    Returns:
        bool
    '''
    from ..models.glyphs import MathTextGlyph
    from ..models.text import MathText

    return _any(all_objs, lambda obj: isinstance(obj, MathTextGlyph | MathText) or _model_requires_mathjax(obj)) or _ext_use_mathjax(all_objs)

def _use_gl(all_objs: set[HasProps]) -> bool:
    ''' Whether a collection of Bokeh objects contains a plot requesting WebGL

    Args:
        objs (seq[HasProps or Document]) :

    Returns:
        bool

    '''
    from ..models.plots import Plot
    return _any(all_objs, lambda obj: isinstance(obj, Plot) and obj.output_backend == "webgl")

def _ext_use_tables(all_objs: set[HasProps]) -> bool:
    from ..models.widgets import TableWidget
    return _query_extensions(all_objs, lambda cls: issubclass(cls, TableWidget))

def _ext_use_widgets(all_objs: set[HasProps]) -> bool:
    from ..models.widgets import Widget
    return _query_extensions(all_objs, lambda cls: issubclass(cls, Widget))

def _ext_use_mathjax(all_objs: set[HasProps]) -> bool:
    from ..models.text import MathText
    return _query_extensions(all_objs, lambda cls: issubclass(cls, MathText))
#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from datetime import date, datetime
from typing import Literal, Sequence

# Bokeh imports
from ..._types import Color, Datetime, Readonly
from ...core.has_props import abstract
from ..formatters import TickFormatter
from .widget import Widget

@abstract
@dataclass(init=False)
class AbstractSlider(Widget):

    orientation: Literal["horizontal", "vertical"] = ...

    title: str | None = ...

    show_value: bool = ...

    direction: Literal["ltr", "rtl"] = ...

    tooltips: bool = ...

    bar_color: Color = ...

@abstract
@dataclass(init=False)
class NumericalSlider(AbstractSlider):

    format: str | TickFormatter = ...

@dataclass
class CategoricalSlider(AbstractSlider):

    categories: Sequence[str] = ...

    value: str = ...

    value_throttled: Readonly[str] = ...

@dataclass
class Slider(NumericalSlider):

    start: float = ...

    end: float = ...

    value: float = ...

    value_throttled: Readonly[float] = ...

    step: float = ...

@dataclass
class RangeSlider(NumericalSlider):

    value: tuple[float, float] = ...

    value_throttled: Readonly[tuple[float, float]] = ...

    start: float = ...

    end: float = ...

    step: float = ...

@dataclass
class DateSlider(NumericalSlider):

    @property
    def value_as_datetime(self) -> datetime | None: ...

    @property
    def value_as_date(self) -> date | None: ...

    value: Datetime = ...

    value_throttled: Readonly[Datetime] = ...

    start: Datetime = ...

    end: Datetime = ...

    step: int = ...

@dataclass
class DateRangeSlider(NumericalSlider):

    @property
    def value_as_datetime(self) -> tuple[datetime, datetime] | None: ...

    @property
    def value_as_date(self) -> tuple[date, date] | None: ...

    value: tuple[Datetime, Datetime] = ...

    value_throttled: Readonly[tuple[Datetime, Datetime]] = ...

    start: Datetime = ...

    end: Datetime = ...

    step: int = ...

@dataclass
class DatetimeRangeSlider(NumericalSlider):

    @property
    def value_as_datetime(self) -> tuple[datetime, datetime] | None: ...

    value: tuple[Datetime, Datetime] = ...

    value_throttled: Readonly[tuple[Datetime, Datetime]] = ...

    start: Datetime = ...

    end: Datetime = ...

    step: int = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.enums import RenderLevelType as RenderLevel
from ...core.has_props import abstract
from ...model import Model
from ..coordinates import CoordinateMapping
from ..dom import DOMNode
from ..ui import Menu, StyledElement, UIElement

@dataclass
class RendererGroup(Model):

    visible: bool = ...

@abstract
@dataclass(init=False)
class Renderer(StyledElement):

    level: RenderLevel = ...

    visible: bool = ...

    coordinates: CoordinateMapping | None = ...

    x_range_name: str = ...

    y_range_name: str = ...

    group: RendererGroup | None = ...

    propagate_hover: bool = ...

    context_menu: Menu | None = ...

@abstract
@dataclass(init=False)
class CompositeRenderer(Renderer):

    renderers: list[Renderer] = ...

    elements: list[UIElement | DOMNode] = ...

@abstract
@dataclass(init=False)
class DataRenderer(Renderer):
    ...

@abstract
@dataclass(init=False)
class GuideRenderer(Renderer):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.has_props import abstract
from .transforms import Transform

@abstract
@dataclass(init=False)
class Scale(Transform):
    ...

@dataclass
class ContinuousScale(Scale):
    ...

@dataclass
class LinearScale(ContinuousScale):
    ...

@dataclass
class LogScale(ContinuousScale):
    ...

@dataclass
class CategoricalScale(Scale):
    ...

@dataclass
class CompositeScale(Scale):

    source_scale: Scale = ...

    target_scale: Scale = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.enums import BuiltinFormatterType as BuiltinFormatter
from ..core.has_props import HasProps, abstract
from ..model import Model, Qualified
from .callbacks import CustomJS
from .css import Styles
from .renderers import RendererGroup
from .tools import CustomJSHover
from .ui import UIElement

@abstract
@dataclass(init=False)
class DOMNode(Model, Qualified):
    ...

@dataclass
class Text(DOMNode):

    content: str = ...

@abstract
@dataclass(init=False)
class DOMElement(DOMNode):

    style: Styles | dict[str, str] = ...

    children: list[str | DOMNode | UIElement] = ...

@dataclass
class Span(DOMElement):
    ...

@dataclass
class Div(DOMElement):
    ...

@dataclass
class Table(DOMElement):
    ...

@dataclass
class TableRow(DOMElement):
    ...

@abstract
@dataclass(init=False)
class Action(Model, Qualified):
    ...

@dataclass
class Template(DOMElement):

    actions: list[Action] = ...

@dataclass
class ToggleGroup(Action):

    groups: list[RendererGroup] = ...

@abstract
@dataclass(init=False)
class Placeholder(DOMElement):
    ...

@dataclass
class ValueOf(Placeholder):

    obj: HasProps = ...

    attr: str = ...

    format: str | None = ...

    formatter: BuiltinFormatter | CustomJS = ...

@dataclass
class Index(Placeholder):
    ...

@dataclass
class ValueRef(Placeholder):

    field: str = ...

    format: str | None = ...

    formatter: BuiltinFormatter | CustomJS | CustomJSHover = ...

@dataclass
class ColorRef(ValueRef):

    hex: bool = ...

    swatch: bool = ...

@dataclass
class HTML(DOMElement):

    html: str | list[str | DOMNode | UIElement] = ...

    refs: list[str | DOMNode | UIElement] = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Sequence

# Bokeh imports
from ...plotting.contour import ContourData
from ..annotations import ContourColorBar
from ..glyph import Glyph
from .glyph_renderer import GlyphRenderer
from .renderer import DataRenderer

@dataclass
class ContourRenderer(DataRenderer):

    line_renderer: GlyphRenderer[Glyph] = ...

    fill_renderer: GlyphRenderer[Glyph] = ...

    levels: Sequence[float] = ...

    def set_data(self, data: ContourData) -> None: ...

    def construct_color_bar(self, **kwargs: Any) -> ContourColorBar: ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal

# Bokeh imports
from ...core.has_props import abstract
from .buttons import ButtonLike
from .widget import Widget

@abstract
@dataclass(init=False)
class AbstractGroup(Widget):

    labels: list[str] = ...

@abstract
@dataclass(init=False)
class ToggleButtonGroup(AbstractGroup, ButtonLike):

    orientation: Literal["horizontal", "vertical"] = ...

@abstract
@dataclass(init=False)
class ToggleInputGroup(AbstractGroup):

    inline: bool = ...

@dataclass
class CheckboxGroup(ToggleInputGroup):

    active: list[int] = ...

@dataclass
class RadioGroup(ToggleInputGroup):

    active: int | None = ...

@dataclass
class CheckboxButtonGroup(ToggleButtonGroup):

    active: list[int] = ...

@dataclass
class RadioButtonGroup(ToggleButtonGroup):

    active: int | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.enums import AutoType as Auto
from ...core.property_aliases import Anchor
from ..nodes import Coordinate, Node
from .panes import Pane

@dataclass
class Panel(Pane):

    position: Coordinate = ...

    anchor: Anchor = ...

    width: Auto | int | Node = ...

    height: Auto | int | Node = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..model import Model
from .ranges import Range
from .scales import Scale

@dataclass
class CoordinateMapping(Model):

    x_source: Range = ...

    y_source: Range = ...

    x_scale: Scale = ...

    y_scale: Scale = ...

    x_target: Range = ...

    y_target: Range = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..._types import Color, FontSize
from ...core.enums import ToolIconType as ToolIcon
from ...core.has_props import abstract
from .ui_element import UIElement

@abstract
@dataclass(init=False)
class Icon(UIElement):

    size: int | FontSize = ...

@dataclass
class BuiltinIcon(Icon):

    icon_name: ToolIcon | str = ...

    color: Color = ...

@dataclass
class SVGIcon(Icon):

    svg: str = ...

@dataclass
class TablerIcon(Icon):

    icon_name: str = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import re
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Iterator,
    Sequence,
)
from weakref import WeakKeyDictionary

# Bokeh imports
from ..core.types import ID
from ..document.document import Document
from ..model import Model, collect_models
from ..settings import settings
from ..themes.theme import Theme
from ..util.dataclasses import dataclass, field
from ..util.serialization import (
    make_globally_unique_css_safe_id,
    make_globally_unique_id,
)

if TYPE_CHECKING:
    from ..document.document import DocJson

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'contains_tex_string',
    'FromCurdoc',
    'is_tex_string',
    'OutputDocumentFor',
    'RenderItem',
    'RenderRoot',
    'RenderRoots',
    'standalone_docs_json',
    'standalone_docs_json_and_render_items',
    'submodel_has_python_callbacks',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

class FromCurdoc:
    ''' This class merely provides a non-None default value for ``theme``
    arguments, since ``None`` itself is a meaningful value for users to pass.

    '''
    pass

@contextmanager
def OutputDocumentFor(objs: Sequence[Model], apply_theme: Theme | type[FromCurdoc] | None = None,
        always_new: bool = False) -> Iterator[Document]:
    ''' Find or create a (possibly temporary) Document to use for serializing
    Bokeh content.

    Typical usage is similar to:

    .. code-block:: python

         with OutputDocumentFor(models):
            (docs_json, [render_item]) = standalone_docs_json_and_render_items(models)

    Inside the context manager, the models will be considered to be part of a single
    Document, with any theme specified, which can thus be serialized as a unit. Where
    possible, OutputDocumentFor attempts to use an existing Document. However, this is
    not possible in three cases:

    * If passed a series of models that have no Document at all, a new Document will
      be created, and all the models will be added as roots. After the context manager
      exits, the new Document will continue to be the models' document.

    * If passed a subset of Document.roots, then OutputDocumentFor temporarily "re-homes"
      the models in a new bare Document that is only available inside the context manager.

    * If passed a list of models that have different documents, then OutputDocumentFor
      temporarily "re-homes" the models in a new bare Document that is only available
      inside the context manager.

    OutputDocumentFor will also perform document validation before yielding, if
    ``settings.perform_document_validation()`` is True.


        objs (seq[Model]) :
            a sequence of Models that will be serialized, and need a common document

        apply_theme (Theme or FromCurdoc or None, optional):
            Sets the theme for the doc while inside this context manager. (default: None)

            If None, use whatever theme is on the document that is found or created

            If FromCurdoc, use curdoc().theme, restoring any previous theme afterwards

            If a Theme instance, use that theme, restoring any previous theme afterwards

        always_new (bool, optional) :
            Always return a new document, even in cases where it is otherwise possible
            to use an existing document on models.

    Yields:
        Document

    '''
    # Note: Comms handling relies on the fact that the new_doc returned
    # has models with the same IDs as they were started with

    if not isinstance(objs, Sequence) or len(objs) == 0 or not all(isinstance(x, Model) for x in objs):
        raise ValueError("OutputDocumentFor expects a non-empty sequence of Models")

    def finish() -> None:
        pass

    docs = {obj.document for obj in objs if obj.document is not None}

    if always_new:
        def finish() -> None:
            _dispose_temp_doc(objs)
        doc = _create_temp_doc(objs)
    else:
        if len(docs) == 0:
            doc = _new_doc()
            for model in objs:
                doc.add_root(model)

        # handle a single shared document
        elif len(docs) == 1:
            doc = docs.pop()

            # we are not using all the roots, make a quick clone for outputting purposes
            if set(objs) != set(doc.roots):
                def finish() -> None:
                    _dispose_temp_doc(objs)
                doc = _create_temp_doc(objs)

            # we are using all the roots of a single doc, just use doc as-is
            pass  # lgtm [py/unnecessary-pass]

        # models have mixed docs, just make a quick clone
        else:
            def finish():
                _dispose_temp_doc(objs)
            doc = _create_temp_doc(objs)

    if settings.perform_document_validation():
        doc.validate()

    _set_temp_theme(doc, apply_theme)
    yield doc
    _unset_temp_theme(doc)

    finish()


class RenderItem:
    def __init__(self, docid: ID | None = None, token: str | None = None, elementid: ID | None = None,
            roots: list[Model] | dict[Model, ID] | None = None, use_for_title: bool | None = None):

        if (docid is None and token is None) or (docid is not None and token is not None):
            raise ValueError("either docid or sessionid must be provided")

        if roots is None:
            roots = dict()
        elif isinstance(roots, list):
            roots = {root: make_globally_unique_id() for root in roots}

        self.docid = docid
        self.token = token
        self.elementid = elementid
        self.roots = RenderRoots(roots)
        self.use_for_title = use_for_title

    def to_json(self) -> dict[str, Any]:
        json: dict[str, Any] = {}

        if self.docid is not None:
            json["docid"] = self.docid
        else:
            json["token"] = self.token

        if self.elementid is not None:
            json["elementid"] = self.elementid

        if self.roots:
            json["roots"] = self.roots.to_json()
            json["root_ids"] = [root.id for root in self.roots]

        if self.use_for_title is not None:
            json["use_for_title"] = self.use_for_title

        return json

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        else:
            return self.to_json() == other.to_json()


@dataclass
class RenderRoot:
    """ Encapsulate data needed for embedding a Bokeh document root.

    Values for ``name`` or ``tags`` are optional. They may be useful for
    querying a collection of roots to find a specific one to embed.

    """

    #: A unique ID to use for the DOM element
    elementid: ID

    #: The Bokeh model ID for this root
    id: ID = field(compare=False)

    #: An optional user-supplied name for this root
    name: str | None = field(default="", compare=False)

    #: A list of any user-supplied tag values for this root
    tags: list[Any] = field(default_factory=list, compare=False)

    def __post_init__(self):
        # Model.name is nullable, and field() won't enforce the default when name=None
        self.name = self.name or ""


class RenderRoots:
    def __init__(self, roots: dict[Model, ID]) -> None:
        self._roots = roots

    def __iter__(self) -> Iterator[RenderRoot]:
        for i in range(0, len(self)):
            yield self[i]

    def __len__(self):
        return len(self._roots.items())

    def __getitem__(self, key: int | str) -> RenderRoot:
        if isinstance(key, int):
            (root, elementid) = list(self._roots.items())[key]
        else:
            for root, elementid in self._roots.items():
                if root.name == key:
                    break
            else:
                raise ValueError(f"root with {key!r} name not found")

        return RenderRoot(elementid, root.id, root.name, root.tags)

    def __getattr__(self, key: str) -> RenderRoot:
        return self.__getitem__(key)

    def to_json(self) -> dict[ID, ID]:
        return {root.id: elementid for root, elementid in self._roots.items()}

    def __repr__(self) -> str:
        return repr(self._roots)

def standalone_docs_json(models: Sequence[Model | Document]) -> dict[ID, DocJson]:
    '''

    '''
    docs_json, _ = standalone_docs_json_and_render_items(models)
    return docs_json

def standalone_docs_json_and_render_items(models: Model | Document | Sequence[Model | Document], *,
        suppress_callback_warning: bool = False) -> tuple[dict[ID, DocJson], list[RenderItem]]:
    '''

    '''
    if isinstance(models, Model | Document):
        models = [models]

    if not (isinstance(models, Sequence) and all(isinstance(x, Model | Document) for x in models)):
        raise ValueError("Expected a Model, Document, or Sequence of Models or Documents")

    if submodel_has_python_callbacks(models) and not suppress_callback_warning:
        log.warning(_CALLBACKS_WARNING)

    docs: dict[Document, tuple[ID, dict[Model, ID]]] = {}
    for model_or_doc in models:
        if isinstance(model_or_doc, Document):
            model = None
            doc = model_or_doc
        else:
            model = model_or_doc
            doc = model.document

            if doc is None:
                raise ValueError("A Bokeh Model must be part of a Document to render as standalone content")

        if doc not in docs:
            docs[doc] = (make_globally_unique_id(), dict())

        (docid, roots) = docs[doc]

        if model is not None:
            roots[model] = make_globally_unique_css_safe_id()
        else:
            for model in doc.roots:
                roots[model] = make_globally_unique_css_safe_id()

    docs_json: dict[ID, DocJson] = {}
    for doc, (docid, _) in docs.items():
        docs_json[docid] = doc.to_json(deferred=False)

    render_items: list[RenderItem] = []
    for _, (docid, roots) in docs.items():
        render_items.append(RenderItem(docid, roots=roots))

    return (docs_json, render_items)

def submodel_has_python_callbacks(models: Sequence[Model | Document]) -> bool:
    ''' Traverses submodels to check for Python (event) callbacks

    '''
    has_python_callback = False
    for model in collect_models(models):
        if len(model._callbacks) > 0 or len(model._event_callbacks) > 0:
            has_python_callback = True
            break

    return has_python_callback

def is_tex_string(text: str) -> bool:
    ''' Whether a string begins and ends with MathJax default delimiters

    Args:
        text (str): String to check

    Returns:
        bool: True if string begins and ends with delimiters, False if not
    '''
    dollars = r"^\$\$.*?\$\$$"
    braces  = r"^\\\[.*?\\\]$"
    parens  = r"^\\\(.*?\\\)$"

    pat = re.compile(f"{dollars}|{braces}|{parens}", flags=re.S)
    return pat.match(text) is not None

def contains_tex_string(text: str) -> bool:
    ''' Whether a string contains any pair of MathJax default delimiters
    Args:
        text (str): String to check
    Returns:
        bool: True if string contains delimiters, False if not
    '''
    # these are non-greedy
    dollars = r"\$\$.*?\$\$"
    braces  = r"\\\[.*?\\\]"
    parens  = r"\\\(.*?\\\)"

    pat = re.compile(f"{dollars}|{braces}|{parens}", flags=re.S)
    return pat.search(text) is not None

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

_CALLBACKS_WARNING = """
You are generating standalone HTML/JS output, but trying to use real Python
callbacks (i.e. with on_change or on_event). This combination cannot work.

Only JavaScript callbacks may be used with standalone output. For more
information on JavaScript callbacks with Bokeh, see:

    https://docs.bokeh.org/en/latest/docs/user_guide/interaction/js_callbacks.html

Alternatively, to use real Python callbacks, a Bokeh server application may
be used. For more information on building and running Bokeh applications, see:

    https://docs.bokeh.org/en/latest/docs/user_guide/server.html
"""

def _new_doc() -> Document:
    # TODO: embed APIs need to actually respect the existing document's
    # configuration, but for now this is better than nothing.
    from ..io import curdoc
    doc = Document()
    callbacks = curdoc().callbacks._js_event_callbacks
    doc.callbacks._js_event_callbacks.update(callbacks)
    return doc

def _create_temp_doc(models: Sequence[Model]) -> Document:
    doc = _new_doc()
    for m in models:
        doc.models[m.id] = m
        m._temp_document = doc
        for ref in m.references():
            doc.models[ref.id] = ref
            ref._temp_document = doc
    doc._roots = list(models)
    return doc

def _dispose_temp_doc(models: Sequence[Model]) -> None:
    for m in models:
        m._temp_document = None
        for ref in m.references():
            ref._temp_document = None

_themes: WeakKeyDictionary[Document, Theme] = WeakKeyDictionary()

def _set_temp_theme(doc: Document, apply_theme: Theme | type[FromCurdoc] | None) -> None:
    _themes[doc] = doc.theme
    if apply_theme is FromCurdoc:
        from ..io import curdoc
        doc.theme = curdoc().theme
    elif isinstance(apply_theme, Theme):
        doc.theme = apply_theme

def _unset_temp_theme(doc: Document) -> None:
    if doc not in _themes:
        return
    doc.theme = _themes[doc]
    del _themes[doc]

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from typing import (
    Generic,
    Literal,
    NotRequired,
    TypeAlias,
    TypedDict,
    TypeVar,
)

# Bokeh imports
from .._types import NonNegative
from ..core.enums import (
    AlignType as Align,
    AnchorType as Anchor_,
    AutoType as Auto,
    HAlignType as HAlign,
    ToolIconType as ToolIcon,
    VAlignType as VAlign,
)
from ..core.property.bases import Property
from ..core.property.visual import ImageType as Image

CSSLengthType: TypeAlias = str
CSSLength: TypeAlias = Property[CSSLengthType]     # 10px 1.2em, etc.

CSSClassType: TypeAlias = str
CSSClass: TypeAlias = Property[CSSClassType]       # ^\..*$

CSSVariableType: TypeAlias = str
CSSVariable: TypeAlias = Property[CSSVariableType] # ^--.*$

DataImageType: TypeAlias = str
DataImage: TypeAlias = Property[DataImageType]     # ^data:image.*$

# Image has to be first! see #12775, temporary fix
IconLikeType: TypeAlias = Image | ToolIcon | CSSClass | CSSVariable | DataImage
IconLike: TypeAlias = Property[IconLikeType]

T = TypeVar("T")

class XY(TypedDict, Generic[T]):
    x: NotRequired[T]
    y: NotRequired[T]

class LRTB(TypedDict, Generic[T]):
    left: NotRequired[T]
    right: NotRequired[T]
    top: NotRequired[T]
    bottom: NotRequired[T]

class Corners(TypedDict, Generic[T]):
    top_left: NotRequired[T]
    top_right: NotRequired[T]
    bottom_right: NotRequired[T]
    bottom_left: NotRequired[T]

PixelsType: TypeAlias = NonNegative[int]
Pixels: TypeAlias = Property[PixelsType]

HAnchorType: TypeAlias = Align | HAlign | float
HAnchor: TypeAlias = Property[HAnchorType]

VAnchorType: TypeAlias = Align | VAlign | float
VAnchor: TypeAlias = Property[VAnchorType]

AnchorType: TypeAlias = Anchor_ | tuple[HAnchor, VAnchor]
Anchor: TypeAlias = Property[AnchorType]

AutoAnchorType: TypeAlias = Auto | Anchor | tuple[Auto | HAnchor, Auto | VAnchor]
AutoAnchor: TypeAlias = Property[AutoAnchorType]

TextAnchorType: TypeAlias = Anchor | Auto
TextAnchor: TypeAlias = Property[TextAnchorType]

BorderRadiusType: TypeAlias = Pixels | tuple[Pixels, Pixels, Pixels, Pixels] | Corners[Pixels]
BorderRadius: TypeAlias = Property[BorderRadiusType]

PaddingType: TypeAlias = Pixels | tuple[Pixels, Pixels] | XY[Pixels] | tuple[Pixels, Pixels, Pixels, Pixels] | Corners[Pixels]
Padding: TypeAlias = Property[PaddingType]

GridSpacingType: TypeAlias = Pixels | tuple[Pixels, Pixels]
GridSpacing: TypeAlias = Property[GridSpacingType]

TrackAlignType: TypeAlias = Literal["start", "center", "end", "auto"]
TrackAlign: TypeAlias = Property[TrackAlignType]

TrackSizeType: TypeAlias = str
TrackSize: TypeAlias = Property[TrackSizeType]

class FullTrackSize(TypedDict):
    size: NotRequired[TrackSize]
    align: NotRequired[TrackAlign]

TrackSizingType: TypeAlias = TrackSize | FullTrackSize
TrackSizing: TypeAlias = Property[TrackSizingType]

TracksSizingType: TypeAlias = TrackSizing | list[TrackSizing] | dict[int, TrackSizing]
TracksSizing: TypeAlias = Property[TracksSizingType]



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ..core.properties import (
    AnyRef,
    Bool,
    Dict,
    Float,
    Int,
    Nullable,
    Override,
    Required,
    String,
)
from ..model import Model

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'TileSource',
    'MercatorTileSource',
    'TMSTileSource',
    'WMTSTileSource',
    'QUADKEYTileSource',
    'BBoxTileSource',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class TileSource(Model):
    ''' A base class for all tile source types.

    In general, tile sources are used as a required input for ``TileRenderer``.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    _args = ('url', 'tile_size', 'min_zoom', 'max_zoom', 'extra_url_vars')

    url = String("", help="""
    Tile service url e.g., http://c.tile.openstreetmap.org/{Z}/{X}/{Y}.png
    """)

    tile_size = Int(default=256, help="""
    Tile size in pixels (e.g. 256)
    """)

    min_zoom = Int(default=0, help="""
    A minimum zoom level for the tile layer. This is the most zoomed-out level.
    """)

    max_zoom = Int(default=30, help="""
    A maximum zoom level for the tile layer. This is the most zoomed-in level.
    """)

    extra_url_vars = Dict(String, AnyRef, help="""
    A dictionary that maps url variable template keys to values.

    These variables are useful for parts of tile urls which do not change from
    tile to tile (e.g. server host name, or layer name).
    """)

    attribution = String("", help="""
    Data provider attribution content. This can include HTML content.
    """)

    x_origin_offset = Required(Float, help="""
    An x-offset in plot coordinates
    """)

    y_origin_offset = Required(Float, help="""
    A y-offset in plot coordinates
    """)

    initial_resolution = Nullable(Float, help="""
    Resolution (plot_units / pixels) of minimum zoom level of tileset
    projection. None to auto-compute.
    """)

class MercatorTileSource(TileSource):
    ''' A base class for Mercator tile services (e.g. ``WMTSTileSource``).

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    _args = ('url', 'tile_size', 'min_zoom', 'max_zoom', 'x_origin_offset', 'y_origin_offset', 'extra_url_vars', 'initial_resolution')

    x_origin_offset = Override(default=20037508.34)

    y_origin_offset = Override(default=20037508.34)

    initial_resolution = Override(default=156543.03392804097)

    snap_to_zoom = Bool(default=False, help="""
    Forces initial extents to snap to the closest larger zoom level.""")

    wrap_around = Bool(default=True, help="""
    Enables continuous horizontal panning by wrapping the x-axis based on
    bounds of map.

    .. note::
        Axis coordinates are not wrapped. To toggle axis label visibility,
        use ``plot.axis.visible = False``.

    """)

class TMSTileSource(MercatorTileSource):
    ''' Contains tile config info and provides urls for tiles based on a
    templated url e.g. ``http://your.tms.server.host/{Z}/{X}/{Y}.png``. The
    defining feature of TMS is the tile-origin in located at the bottom-left.

    ``TMSTileSource`` can also be helpful in implementing tile renderers for
    custom tile sets, including non-spatial datasets.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class WMTSTileSource(MercatorTileSource):
    ''' Behaves much like ``TMSTileSource`` but has its tile-origin in the
    top-left.

    This is the most common used tile source for web mapping applications.
    Such companies as Google, MapQuest, Stadia, Esri, and OpenStreetMap provide
    service which use the WMTS specification e.g. ``http://c.tile.openstreetmap.org/{Z}/{X}/{Y}.png``.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class QUADKEYTileSource(MercatorTileSource):
    ''' Has the same tile origin as the ``WMTSTileSource`` but requests tiles using
    a `quadkey` argument instead of X, Y, Z e.g.
    ``http://your.quadkey.tile.host/{Q}.png``

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class BBoxTileSource(MercatorTileSource):
    ''' Has the same default tile origin as the ``WMTSTileSource`` but requested
    tiles use a ``{XMIN}``, ``{YMIN}``, ``{XMAX}``, ``{YMAX}`` e.g.
    ``http://your.custom.tile.service?bbox={XMIN},{YMIN},{XMAX},{YMAX}``.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    use_latlon = Bool(default=False, help="""
    Flag which indicates option to output ``{XMIN}``, ``{YMIN}``, ``{XMAX}``, ``{YMAX}`` in meters or latitude and longitude.
    """)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.has_props import abstract
from .widget import Widget

@abstract
@dataclass(init=False)
class Markup(Widget):

    text: str = ...

    disable_math: bool = ...

@dataclass
class Paragraph(Markup):
    ...

@dataclass
class Div(Markup):

    render_as_text: bool = ...

@dataclass
class PreText(Paragraph):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from .._types import JSON, Bytes
from ..core.enums import MapTypeType as MapType
from ..core.has_props import abstract
from ..model import Model
from .plots import Plot

@abstract
@dataclass(init=False)
class MapOptions(Model):

    lat: float = ...

    lng: float = ...

    zoom: int = ...

@abstract
@dataclass(init=False)
class MapPlot(Plot):
    ...

@dataclass
class GMapOptions(MapOptions):

    map_type: MapType = ...

    scale_control: bool = ...

    styles: JSON | None = ...

    tilt: int = ...

@dataclass
class GMapPlot(MapPlot):

    map_options: GMapOptions = ...

    api_key: Bytes | str = ...

    api_version: str = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Literal,
    NotRequired,
    Sequence,
    TypedDict,
)

# Bokeh imports
from .._types import (
    Alpha,
    Color,
    Date,
    Datetime,
    NonNegative,
    Percent,
)
from ..core.enums import (
    AnchorType as Anchor,
    AutoType as Auto,
    DimensionsType as Dimensions,
    DimensionType as Dimension,
    PanDirectionType as PanDirection,
    RegionSelectionModeType as RegionSelectionMode,
    SelectionModeType as SelectionMode,
    ToolNameType as ToolName,
    TooltipAttachmentType as TooltipAttachment,
    TooltipFieldFormatterType as TooltipFieldFormatter,
)
from ..core.has_props import abstract
from ..core.property_aliases import IconLikeType as IconLike
from ..model import Model
from .annotations import BoxAnnotation, PolyAnnotation, Span
from .callbacks import Callback
from .dom import DOMElement
from .glyphs import (
    HStrip,
    Line,
    LRTBGlyph,
    MultiLine,
    Patches,
    Rect,
    VStrip,
    XYGlyph,
)
from .misc.group_by import GroupBy
from .ranges import Range
from .renderers import DataRenderer, GlyphRenderer
from .ui import Menu, UIElement

class Modifiers(TypedDict):
    shift: NotRequired[bool]
    ctrl: NotRequired[bool]
    alt: NotRequired[bool]

@abstract
@dataclass(init=False)
class Tool(Model):

    icon: IconLike | None = ...

    description: str | None = ...

    visible: bool = ...

    group: str | bool = ...

    _known_aliases: ClassVar[dict[str, Callable[[], Tool]]]

    @classmethod
    def from_string(cls, name: str) -> Tool: ...

    @classmethod
    def register_alias(cls, name: str, constructor: Callable[[], Tool]) -> None: ...

@dataclass
class ToolProxy(Model):

    tools: list[Tool] = ...

    active: bool = ...

    disabled: bool = ...

@abstract
@dataclass(init=False)
class ActionTool(Tool):
    ...

@abstract
@dataclass(init=False)
class PlotActionTool(ActionTool):
    ...

@abstract
@dataclass(init=False)
class GestureTool(Tool):
    ...

@abstract
@dataclass(init=False)
class Drag(GestureTool):
    ...

@abstract
@dataclass(init=False)
class Scroll(GestureTool):
    ...

@abstract
@dataclass(init=False)
class Tap(GestureTool):
    ...

@abstract
@dataclass(init=False)
class SelectTool(GestureTool):

    renderers: Auto | list[DataRenderer] = ...

@abstract
@dataclass(init=False)
class RegionSelectTool(SelectTool):

    mode: RegionSelectionMode = ...

    continuous: bool = ...

    select_every_mousemove: bool = ...

    persistent: bool = ...

    greedy: bool = ...

@abstract
@dataclass(init=False)
class InspectTool(GestureTool):

    toggleable: bool = ...

@dataclass
class Toolbar(UIElement):

    tools: list[Tool | ToolProxy] = ...

    logo: Literal["normal", "grey"] | None = ...

    autohide: bool = ...

    group: bool = ...

    group_types: list[ToolName] = ...

    active_drag: Auto | Drag | ToolProxy | None = ...

    active_inspect: Auto | InspectTool | ToolProxy | Sequence[InspectTool] | None = ...

    active_scroll: Auto | Scroll | ToolProxy | None = ...

    active_tap: Auto | Tap | ToolProxy | None = ...

    active_multi: Auto | GestureTool | ToolProxy | None = ...

@dataclass
class ToolMenu(Menu):

    toolbar: Toolbar = ...

@dataclass
class PanTool(Drag):

    dimensions: Dimensions = ...

@dataclass
class ClickPanTool(PlotActionTool):

    direction: PanDirection = ...

    factor: Percent = ...

@dataclass
class RangeTool(Tool):

    x_range: Range | None = ...

    y_range: Range | None = ...

    x_interaction: bool = ...

    y_interaction: bool = ...

    overlay: BoxAnnotation = ...

    start_gesture: Literal["pan", "tap", "none"] = ...

@dataclass
class WheelPanTool(Scroll):

    dimension: Dimension = ...

    modifiers: Modifiers | str = ...

@dataclass
class WheelZoomTool(Scroll):

    dimensions: Dimensions = ...

    renderers: Auto | list[DataRenderer] = ...

    level: NonNegative[int] = ...

    hit_test: bool = ...

    hit_test_mode: Literal["point", "hline", "vline"] = ...

    hit_test_behavior: GroupBy | Literal["only_hit", "group_by_name"] | list[list[DataRenderer]] = ...

    maintain_focus: bool = ...

    zoom_on_axis: bool = ...

    zoom_together: Literal["none", "cross", "all"] = ...

    speed: float = ...

    modifiers: Modifiers | str = ...

@dataclass
class CustomAction(ActionTool):

    active: bool = ...

    disabled: bool = ...

    callback: Callback | None = ...

    active_callback: Callback | Auto | None = ...

@dataclass
class SaveTool(ActionTool):

    filename: str | None = ...

@dataclass
class CopyTool(ActionTool):
    ...

@dataclass
class ResetTool(PlotActionTool):
    ...

@dataclass
class TapTool(Tap, SelectTool):

    mode: SelectionMode = ...

    behavior: Literal["select", "inspect"] = ...

    gesture: Literal["tap", "doubletap"] = ...

    modifiers: Modifiers | str = ...

    callback: Callback | None = ...

@dataclass
class CrosshairTool(InspectTool):

    overlay: Auto | Span | tuple[Span, Span] = ...

    dimensions: Dimensions = ...

    line_color: Color = ...

    line_alpha: Alpha = ...

    line_width: float = ...

@dataclass
class BoxZoomTool(Drag):

    dimensions: Dimensions | Auto = ...

    overlay: BoxAnnotation = ...

    match_aspect: bool = ...

    origin: Literal["corner", "center"] = ...

@abstract
@dataclass(init=False)
class ZoomBaseTool(PlotActionTool):

    renderers: Auto | list[DataRenderer] = ...

    dimensions: Dimensions = ...

    factor: Percent = ...

    level: NonNegative[int] = ...

@dataclass
class ZoomInTool(ZoomBaseTool):
    ...

@dataclass
class ZoomOutTool(ZoomBaseTool):

    maintain_focus: bool = ...

@dataclass
class BoxSelectTool(Drag, RegionSelectTool):

    dimensions: Dimensions = ...

    overlay: BoxAnnotation = ...

    origin: Literal["corner", "center"] = ...

@dataclass
class LassoSelectTool(Drag, RegionSelectTool):

    overlay: PolyAnnotation = ...

@dataclass
class PolySelectTool(Tap, RegionSelectTool):

    overlay: PolyAnnotation = ...

@dataclass
class CustomJSHover(Model):

    args: dict[str, Any] = ...

    code: str = ...

@dataclass
class HoverTool(InspectTool):

    renderers: Auto | list[DataRenderer] = ...

    callback: Callback | None = ...

    tooltips: None | DOMElement | str | list[tuple[str, str]] | dict[str, str] = ...

    formatters: dict[str, TooltipFieldFormatter | CustomJSHover] = ...

    mode: Literal["mouse", "hline", "vline"] = ...

    muted_policy: Literal["show", "ignore"] = ...

    point_policy: Literal["snap_to_data", "follow_mouse", "none"] = ...

    line_policy: Literal["prev", "next", "nearest", "interp", "none"] = ...

    anchor: Anchor = ...

    attachment: TooltipAttachment = ...

    show_arrow: bool = ...

@dataclass
class HelpTool(ActionTool):

    redirect: str = ...

@dataclass
class ExamineTool(ActionTool):
    ...

@dataclass
class FullscreenTool(ActionTool):
    ...

@dataclass
class UndoTool(PlotActionTool):
    ...

@dataclass
class RedoTool(PlotActionTool):
    ...

@abstract
@dataclass(init=False)
class EditTool(GestureTool):

    default_overrides: dict[str, Any] = ...

    empty_value: bool | int | float | Date | Datetime | Color | str = ...

@abstract
@dataclass(init=False)
class PolyTool(EditTool):

    vertex_renderer: GlyphRenderer[XYGlyph] | None = ...

@dataclass
class BoxEditTool(EditTool, Drag, Tap):

    renderers: list[GlyphRenderer[LRTBGlyph | Rect | HStrip | VStrip]] = ...

    dimensions: Dimensions = ...

    num_objects: int = ...

@dataclass
class PointDrawTool(EditTool, Drag, Tap):

    renderers: list[GlyphRenderer[XYGlyph]] = ...

    add: bool = ...

    drag: bool = ...

    num_objects: int = ...

@dataclass
class PolyDrawTool(PolyTool, Drag, Tap):

    renderers: list[GlyphRenderer[MultiLine | Patches]] = ...

    drag: bool = ...

    num_objects: int = ...

@dataclass
class FreehandDrawTool(EditTool, Drag, Tap):

    renderers: list[GlyphRenderer[MultiLine | Patches]] = ...

    num_objects: int = ...

@dataclass
class PolyEditTool(PolyTool, Drag, Tap):

    renderers: list[GlyphRenderer[MultiLine | Patches]] = ...

@dataclass
class LineEditTool(EditTool, Drag, Tap):

    renderers: list[GlyphRenderer[Line]] = ...

    intersection_renderer: GlyphRenderer[Line] = ...

    dimensions: Dimensions = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..._specs import AngleSpec, NullStringSpec, NumberSpec
from ..._types import Angle, Coordinate, TextLike
from ...core.enums import (
    AngleUnitsType as AngleUnits,
    CoordinateUnitsType as CoordinateUnits,
    DirectionType as Direction,
    TextAlignType as TextAlign,
    VerticalAlignType as VerticalAlign,
)
from ...core.has_props import abstract
from ...core.property_aliases import BorderRadius, Padding, TextAnchor
from ...core.property_mixins import (
    BackgroundFillProps,
    BackgroundHatchProps,
    BorderLineProps,
    ScalarBackgroundFillProps,
    ScalarBackgroundHatchProps,
    ScalarBorderLineProps,
    ScalarTextProps,
    TextProps,
)
from .annotation import Annotation, DataAnnotation

@abstract
@dataclass(init=False)
class TextAnnotation(Annotation, ScalarTextProps, ScalarBackgroundFillProps, ScalarBackgroundHatchProps, ScalarBorderLineProps):

    text: TextLike = ...

    padding: Padding = ...

    border_radius: BorderRadius = ...

@dataclass
class Label(TextAnnotation):

    anchor: TextAnchor = ...

    x: Coordinate = ...

    y: Coordinate = ...

    x_units: CoordinateUnits = ...

    y_units: CoordinateUnits = ...

    x_offset: float = ...

    y_offset: float = ...

    angle: Angle = ...

    angle_units: AngleUnits = ...

    direction: Direction = ...

    editable: bool = ...

@dataclass
class LabelSet(DataAnnotation, TextProps, BackgroundFillProps, BackgroundHatchProps, BorderLineProps):

    x: NumberSpec = ...

    x_units: CoordinateUnits = ...

    y: NumberSpec = ...

    y_units: CoordinateUnits = ...

    text: NullStringSpec = ...

    angle: AngleSpec = ...

    x_offset: NumberSpec = ...

    y_offset: NumberSpec = ...

@dataclass
class Title(TextAnnotation):

    vertical_align: VerticalAlign = ...

    align: TextAlign = ...

    standoff: float = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from .._specs import (
    AngleSpec,
    DataSpec,
    DistanceSpec,
    FloatSpec,
    MarkerSpec,
    NonNegative,
    NullDistanceSpec,
    NumberSpec,
    SizeSpec,
    StringSpec,
)
from ..core.enums import (
    DirectionType as Direction,
    HexTileOrientationType as HexTileOrientation,
    ImageOriginType as ImageOrigin,
    OutlineShapeNameType as OutlineShapeName,
    PaletteType as Palette,
    RadiusDimensionType as RadiusDimension,
    StepModeType as StepMode,
    TeXDisplayType as TeXDisplay,
)
from ..core.has_props import abstract
from ..core.property_aliases import (
    Anchor,
    BorderRadius,
    Padding,
    TextAnchor,
)
from ..core.property_mixins import (
    BackgroundFillProps,
    BackgroundHatchProps,
    BorderLineProps,
    FillProps,
    HatchProps,
    ImageProps,
    LineProps,
    ScalarFillProps,
    ScalarHatchProps,
    ScalarLineProps,
    TextProps,
)
from .callbacks import CustomJS
from .glyph import (
    ConnectedXYGlyph,
    Glyph,
    RadialGlyph,
    XYGlyph,
)
from .mappers import ColorMapper, StackColorMapper

@abstract
@dataclass(init=False)
class Marker(XYGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    hit_dilation: NonNegative[float] = ...

    size: SizeSpec = ...

    angle: AngleSpec = ...

@abstract
@dataclass(init=False)
class LRTBGlyph(Glyph, LineProps, FillProps, HatchProps):

    border_radius: BorderRadius = ...

@dataclass
class AnnularWedge(XYGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    inner_radius: DistanceSpec = ...

    outer_radius: DistanceSpec = ...

    start_angle: AngleSpec = ...

    end_angle: AngleSpec = ...

    direction: Direction = ...

@dataclass
class Annulus(XYGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    inner_radius: DistanceSpec = ...

    outer_radius: DistanceSpec = ...

@dataclass
class Arc(XYGlyph, LineProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    radius: DistanceSpec = ...

    start_angle: AngleSpec = ...

    end_angle: AngleSpec = ...

    direction: Direction = ...

@dataclass
class Bezier(Glyph, LineProps):

    x0: NumberSpec = ...

    y0: NumberSpec = ...

    x1: NumberSpec = ...

    y1: NumberSpec = ...

    cx0: NumberSpec = ...

    cy0: NumberSpec = ...

    cx1: NumberSpec = ...

    cy1: NumberSpec = ...

@dataclass
class Block(LRTBGlyph):

    x: NumberSpec = ...

    y: NumberSpec = ...

    width: DistanceSpec = ...

    height: DistanceSpec = ...

@dataclass
class Circle(RadialGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    radius: DistanceSpec = ...

    radius_dimension: RadiusDimension = ...

    hit_dilation: NonNegative[float] = ...

@dataclass
class Ellipse(XYGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    width: DistanceSpec = ...

    height: DistanceSpec = ...

    angle: AngleSpec = ...

@dataclass
class HArea(Glyph, ScalarFillProps, HatchProps):

    x1: NumberSpec = ...

    x2: NumberSpec = ...

    y: NumberSpec = ...

@dataclass
class HAreaStep(Glyph, ScalarFillProps, HatchProps):

    x1: NumberSpec = ...

    x2: NumberSpec = ...

    y: NumberSpec = ...

    step_mode: StepMode = ...

@dataclass
class HBar(LRTBGlyph):

    y: NumberSpec = ...

    height: DistanceSpec = ...

    left: NumberSpec = ...

    right: NumberSpec = ...

@dataclass
class HSpan(Glyph, LineProps):

    y: NumberSpec = ...

@dataclass
class HStrip(Glyph, LineProps, FillProps, HatchProps):

    y0: NumberSpec = ...

    y1: NumberSpec = ...

@dataclass
class HexTile(Glyph, LineProps, FillProps, HatchProps):

    size: float = ...

    aspect_scale: float = ...

    r: NumberSpec = ...

    q: NumberSpec = ...

    scale: NumberSpec = ...

    orientation: HexTileOrientation = ...

@abstract
@dataclass(init=False)
class ImageBase(XYGlyph, ImageProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    dw: DistanceSpec = ...

    dh: DistanceSpec = ...

    dilate: bool = ...

    origin: ImageOrigin = ...

    anchor: Anchor = ...

@dataclass
class Image(ImageBase):

    image: NumberSpec = ...

    color_mapper: ColorMapper | Palette = ...

@dataclass
class ImageRGBA(ImageBase):

    image: NumberSpec = ...

@dataclass
class ImageStack(ImageBase):

    image: NumberSpec = ...

    color_mapper: StackColorMapper = ...

@dataclass
class ImageURL(XYGlyph):

    url: StringSpec = ...

    x: NumberSpec = ...

    y: NumberSpec = ...

    w: NullDistanceSpec = ...

    h: NullDistanceSpec = ...

    angle: AngleSpec = ...

    global_alpha: NumberSpec = ...

    dilate: bool = ...

    anchor: Anchor = ...

    retry_attempts: int = ...

    retry_timeout: int = ...

@dataclass
class Line(ConnectedXYGlyph, ScalarLineProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

@dataclass
class MultiLine(Glyph, LineProps):

    xs: NumberSpec = ...

    ys: NumberSpec = ...

@dataclass
class MultiPolygons(Glyph, LineProps, FillProps, HatchProps):

    xs: NumberSpec = ...

    ys: NumberSpec = ...

@dataclass
class Ngon(RadialGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    radius: DistanceSpec = ...

    angle: AngleSpec = ...

    n: NumberSpec = ...

    radius_dimension: RadiusDimension = ...

@dataclass
class Patch(ConnectedXYGlyph, ScalarLineProps, ScalarFillProps, ScalarHatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

@dataclass
class Patches(Glyph, LineProps, FillProps, HatchProps):

    xs: NumberSpec = ...

    ys: NumberSpec = ...

@dataclass
class Quad(LRTBGlyph):

    left: NumberSpec = ...

    right: NumberSpec = ...

    bottom: NumberSpec = ...

    top: NumberSpec = ...

@dataclass
class Quadratic(Glyph, LineProps):

    x0: NumberSpec = ...

    y0: NumberSpec = ...

    x1: NumberSpec = ...

    y1: NumberSpec = ...

    cx: NumberSpec = ...

    cy: NumberSpec = ...

@dataclass
class Ray(XYGlyph, LineProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    angle: AngleSpec = ...

    length: DistanceSpec = ...

@dataclass
class Rect(XYGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    width: DistanceSpec = ...

    height: DistanceSpec = ...

    angle: AngleSpec = ...

    border_radius: BorderRadius = ...

    dilate: bool = ...

@dataclass
class Scatter(Marker):

    marker: MarkerSpec = ...

    defs: dict[str, CustomJS] = ...

@dataclass
class Segment(Glyph, LineProps):

    x0: NumberSpec = ...

    y0: NumberSpec = ...

    x1: NumberSpec = ...

    y1: NumberSpec = ...

@dataclass
class Step(XYGlyph, ScalarLineProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    mode: StepMode = ...

@dataclass
class Text(XYGlyph, TextProps, BackgroundFillProps, BackgroundHatchProps, BorderLineProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    text: StringSpec = ...

    angle: AngleSpec = ...

    x_offset: FloatSpec = ...

    y_offset: FloatSpec = ...

    anchor: DataSpec[TextAnchor] = ...

    padding: Padding = ...

    border_radius: BorderRadius = ...

    outline_shape: DataSpec[OutlineShapeName] = ...

@abstract
@dataclass(init=False)
class MathTextGlyph(Text):
    ...

@dataclass
class MathMLGlyph(MathTextGlyph):
    ...

@dataclass
class TeXGlyph(MathTextGlyph):

    macros: dict[str, str | tuple[str, int]] = ...

    display: TeXDisplay = ...

@dataclass
class VArea(Glyph, ScalarFillProps, HatchProps):

    x: NumberSpec = ...

    y1: NumberSpec = ...

    y2: NumberSpec = ...

@dataclass
class VAreaStep(Glyph, ScalarFillProps, HatchProps):

    x: NumberSpec = ...

    y1: NumberSpec = ...

    y2: NumberSpec = ...

    step_mode: StepMode = ...

@dataclass
class VBar(LRTBGlyph):

    x: NumberSpec = ...

    width: DistanceSpec = ...

    bottom: NumberSpec = ...

    top: NumberSpec = ...

@dataclass
class VSpan(Glyph, LineProps):

    x: NumberSpec = ...

@dataclass
class VStrip(Glyph, LineProps, FillProps, HatchProps):

    x0: NumberSpec = ...

    x1: NumberSpec = ...

@dataclass
class Wedge(XYGlyph, LineProps, FillProps, HatchProps):

    x: NumberSpec = ...

    y: NumberSpec = ...

    radius: DistanceSpec = ...

    start_angle: AngleSpec = ...

    end_angle: AngleSpec = ...

    direction: Direction = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
""" Provide (optional) Pandas properties.

"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import TYPE_CHECKING, Any

# Bokeh imports
from .bases import Property

if TYPE_CHECKING:
    from narwhals.stable.v1.typing import IntoDataFrame, IntoSeries  # noqa: F401
    from pandas import DataFrame  # noqa: F401
    from pandas.core.groupby import GroupBy  # noqa: F401


#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'EagerDataFrame',
    'EagerSeries',
    'PandasDataFrame',
    'PandasGroupBy',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class EagerDataFrame(Property["IntoDataFrame"]):
    """ Accept eager dataframe supported by Narwhals.

    This property only exists to support type validation, e.g. for "accepts"
    clauses. It is not serializable itself, and is not useful to add to
    Bokeh models directly.

    """
    def validate(self, value: Any, detail: bool = True) -> None:
        import narwhals.stable.v1 as nw
        super().validate(value, detail)

        if nw.dependencies.is_into_dataframe(value):
            return # type: ignore[unreachable] # https://github.com/bokeh/bokeh/issues/14342

        msg = "" if not detail else f"expected object convertible to Narwhals DataFrame, got {value!r}"
        raise ValueError(msg)

class EagerSeries(Property["IntoSeries"]):
    """ Accept eager series supported by Narwhals.

    This property only exists to support type validation, e.g. for "accepts"
    clauses. It is not serializable itself, and is not useful to add to
    Bokeh models directly.

    """
    def validate(self, value: Any, detail: bool = True) -> None:
        import narwhals.stable.v1 as nw
        super().validate(value, detail)

        if nw.dependencies.is_into_series(value):
            return # type: ignore[unreachable] # https://github.com/bokeh/bokeh/issues/14342

        msg = "" if not detail else f"expected object convertible to Narwhals Series, got {value!r}"
        raise ValueError(msg)

class PandasDataFrame(Property["DataFrame"]):
    """ Accept Pandas DataFrame values.

    This class is pandas-specific - are more generic one is
    ``EagerDataFrame()``.

    This property only exists to support type validation, e.g. for "accepts"
    clauses. It is not serializable itself, and is not useful to add to
    Bokeh models directly.

    """
    def __init__(self) -> None:
        super().__init__()


    def validate(self, value: Any, detail: bool = True) -> None:
        import pandas as pd
        if isinstance(value, pd.DataFrame):
            return

        msg = "" if not detail else f"expected Pandas DataFrame, got {value!r}"
        raise ValueError(msg)

class PandasGroupBy(Property["GroupBy[Any]"]):
    """ Accept Pandas DataFrame values.

    This property only exists to support type validation, e.g. for "accepts"
    clauses. It is not serializable itself, and is not useful to add to
    Bokeh models directly.

    """

    def validate(self, value: Any, detail: bool = True) -> None:
        super().validate(value, detail)

        from pandas.core.groupby import GroupBy
        if isinstance(value, GroupBy):
            return

        msg = "" if not detail else f"expected Pandas GroupBy, got {value!r}"
        raise ValueError(msg)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Sequence

# Bokeh imports
from ..core.enums import (
    AutoType as Auto,
    ContextWhichType as ContextWhich,
    LatLonType as LatLon,
    LocationType as Location,
    NumeralLanguageType as NumeralLanguage,
    ResolutionTypeType as ResolutionType,
    RoundingFunctionType as RoundingFunction,
)
from ..core.has_props import abstract
from ..model import Model
from .tickers import Ticker

@abstract
@dataclass(init=False)
class TickFormatter(Model):
    ...

@dataclass
class BasicTickFormatter(TickFormatter):

    precision: Auto | int = ...

    use_scientific: bool = ...

    power_limit_high: int = ...

    power_limit_low: int = ...

@dataclass
class MercatorTickFormatter(BasicTickFormatter):

    dimension: LatLon | None = ...

@dataclass
class NumeralTickFormatter(TickFormatter):

    format: str = ...

    language: NumeralLanguage = ...

    rounding: RoundingFunction = ...

@dataclass
class PrintfTickFormatter(TickFormatter):

    format: str = ...

@dataclass
class LogTickFormatter(TickFormatter):

    ticker: Ticker | None = ...

    min_exponent: int = ...

@dataclass
class CategoricalTickFormatter(TickFormatter):
    ...

@dataclass
class CustomJSTickFormatter(TickFormatter):

    args: dict[str, Any] = ...

    code: str = ...

@dataclass
class DatetimeTickFormatter(TickFormatter):

    microseconds: str = ...

    milliseconds: str = ...

    seconds: str = ...

    minsec: str = ...

    minutes: str = ...

    hourmin: str = ...

    hours: str = ...

    days: str = ...

    months: str = ...

    years: str = ...

    strip_leading_zeros: bool | Sequence[ResolutionType] = ...

    boundary_scaling: bool = ...

    hide_repeats: bool = ...

    context: str | DatetimeTickFormatter | None = ...

    context_which: ContextWhich = ...

    context_location: Location = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal, Sequence, TypeAlias

# Bokeh imports
from ..._specs import NullStringSpec
from ..._types import CoordinateLike, TextLike
from ...core.enums import (
    AlignType as Align,
    AlternationPolicyType as AlternationPolicy,
    AnchorType as HVAlign,
    AutoType as Auto,
    HAlignType as HAlign,
    LegendClickPolicyType as LegendClickPolicy,
    LegendLocationType as LegendLocation,
    LocationType as Location,
    OrientationType as Orientation,
    VAlignType as VAlign,
)
from ...core.has_props import abstract
from ...core.property_aliases import AutoAnchor, BorderRadius, Padding
from ...core.property_mixins import (
    ScalarBackgroundFillProps,
    ScalarBackgroundHatchProps,
    ScalarBarLineProps,
    ScalarBorderLineProps,
    ScalarInactiveFillProps,
    ScalarItemBackgroundFillProps,
    ScalarLabelTextProps,
    ScalarMajorLabelTextProps,
    ScalarMajorTickLineProps,
    ScalarMinorTickLineProps,
    ScalarTitleTextProps,
)
from ...model import Model
from ...util.callback_manager import EventCallback as PyEventCallback
from ..callbacks import Callback as JsEventCallback
from ..formatters import TickFormatter
from ..glyph import Glyph
from ..labeling import LabelingPolicy
from ..mappers import ColorMapper
from ..ranges import Range
from ..renderers import GlyphRenderer
from ..tickers import Ticker
from .annotation import Annotation
from .dimensional import Dimensional

@abstract
@dataclass(init=False)
class BaseColorBar(Annotation, ScalarTitleTextProps, ScalarMajorLabelTextProps, ScalarMajorTickLineProps,
                   ScalarMinorTickLineProps, ScalarBarLineProps, ScalarBorderLineProps, ScalarBackgroundFillProps):

    location: HVAlign | tuple[float, float] = ...

    orientation: Orientation | Auto = ...

    height: Auto | int = ...

    width: Auto | int = ...

    scale_alpha: float = ...

    title: TextLike | None = ...

    title_standoff: int = ...

    ticker: Ticker | Auto = ...

    formatter: TickFormatter | Auto = ...

    major_label_overrides: dict[float | str, TextLike] = ...

    major_label_policy: LabelingPolicy = ...

    margin: int = ...

    padding: int = ...

    label_standoff: int = ...

    major_tick_in: int = ...

    major_tick_out: int = ...

    minor_tick_in: int = ...

    minor_tick_out: int = ...

@dataclass
class ColorBar(BaseColorBar):

    color_mapper: ColorMapper = ...

    display_low: float | None = ...

    display_high: float | None = ...

@dataclass
class ContourColorBar(BaseColorBar):

    fill_renderer: GlyphRenderer[Glyph] = ...

    line_renderer: GlyphRenderer[Glyph] = ...

    levels: Sequence[float] = ...

@dataclass
class LegendItem(Model):

    label: NullStringSpec = ...

    renderers: list[GlyphRenderer[Glyph]] = ...

    index: int | None = ...

    visible: bool = ...

@dataclass
class Legend(Annotation, ScalarTitleTextProps, ScalarBorderLineProps, ScalarBackgroundFillProps,
             ScalarItemBackgroundFillProps, ScalarInactiveFillProps, ScalarLabelTextProps):

    location: LegendLocation | tuple[float, float] = ...

    orientation: Orientation = ...

    ncols: int | Auto = ...

    nrows: int | Auto = ...

    title: str | None = ...


    title_location: Location = ...

    title_standoff: int = ...

    click_policy: LegendClickPolicy = ...

    item_background_policy: AlternationPolicy = ...

    label_standoff: int = ...

    label_height: Auto | int = ...

    label_width: int = ...

    glyph_height: int = ...

    glyph_width: int = ...

    margin: int = ...

    padding: Padding = ...

    border_radius: BorderRadius = ...

    spacing: int = ...

    items: list[LegendItem] | list[tuple[str, list[GlyphRenderer[Glyph]]]] = ...

    def on_click(self, handler: PyEventCallback) -> None: ...

    def js_on_click(self, handler: JsEventCallback) -> None: ...

X: TypeAlias = HAlign | float | CoordinateLike
Y: TypeAlias = VAlign | float | CoordinateLike

Position: TypeAlias = HVAlign | tuple[X, Y]
PositionUnits: TypeAlias = Literal["data", "screen", "view", "percent"]

@dataclass
class ScaleBar(Annotation, ScalarBarLineProps, ScalarLabelTextProps, ScalarTitleTextProps,
               ScalarBorderLineProps, ScalarBackgroundFillProps, ScalarBackgroundHatchProps):

    range: Range | Auto = ...

    unit: str = ...

    dimensional: Dimensional = ...

    orientation: Orientation = ...

    location: Position = ...

    x_units: PositionUnits = ...

    y_units: PositionUnits = ...

    anchor: AutoAnchor = ...

    length_sizing: Literal["adaptive", "exact"] = ...

    bar_length: float | int = ...

    bar_length_units: Literal["screen", "data", "percent"] = ...

    margin: int = ...

    padding: int = ...

    label: str = ...

    label_align: Align = ...

    label_location: Location = ...

    label_standoff: int = ...

    title: str = ...

    title_align: Align = ...

    title_location: Location = ...

    title_standoff: int = ...

    ticker: Ticker = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Represent comparisons to be computed on the client (browser) side by
BokehJS. These comparisons may be useful for specifying how DataTable columns
are sorted.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ..core.has_props import abstract
from ..core.properties import (
    AnyRef,
    Bool,
    Dict,
    String,
)
from ..model import Model

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'Comparison',
    'CustomJSCompare',
    'NanCompare',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Comparison(Model):
    ''' Base class for ``Comparison`` models that represent a comparison
    to be carried out on the client-side.

    The JavaScript implementation should implement the following method:

    .. code-block:: typescript

        compute(x: unknown, y: unknown): -1 | 0 | 1 {
            //  compare and return -1, 0, or 1
        }

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class CustomJSCompare(Comparison):
    ''' A client-side comparison performed by evaluating a user-supplied
    JavaScript function. This comparison can be useful for DataTable columns.

    .. warning::
        The explicit purpose of this Bokeh Model is to embed *raw JavaScript
        code* for a browser to execute. If any part of the code is derived
        from untrusted user inputs, then you must take appropriate care to
        sanitize the user input prior to passing to Bokeh.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    args = Dict(String, AnyRef, help="""
    A mapping of names to Python objects. In particular those can be bokeh's models.
    These objects are made available to the callback's code snippet as the values of
    named parameters to the callback. There is no need to manually include the data
    source of the associated glyph renderer, as it is available within the scope of
    the code via `this` keyword (e.g. `this.data` will give access to raw data).
    """)

    code = String(default="", help="""
    A snippet of JavaScript code to execute in the browser. The code is made into
    the body of a generator function and all of the named objects in ``args``
    are available as parameters that the code can use. Must return -1, 0, or 1.
    """)

class NanCompare(Comparison):
    ''' A client-side comparison that can sort NaN values first or last. This
    comparison can be useful for DataTable columns.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    ascending_first = Bool(default=False, help="""
    Whether NaN values should appear first or last in an ascending sort.
    """)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from .._specs import (
    AlphaSpec,
    ColorSpec,
    DashPatternSpec,
    FloatSpec,
    FontSizeSpec,
    FontStyleSpec,
    HatchPatternSpec,
    IntSpec,
    LineCapSpec,
    LineJoinSpec,
    NumberSpec,
    StringSpec,
    TextAlignSpec,
    TextBaselineSpec,
)
from .._types import (
    Alpha,
    Color,
    FontSize,
    Size,
)
from ..models.textures import Texture
from .enums import (
    FontStyleType as FontStyle,
    LineCapType as LineCap,
    LineJoinType as LineJoin,
    TextAlignType as TextAlign,
    TextBaselineType as TextBaseline,
)
from .has_props import HasProps
from .property.visual import DashPatternType as DashPattern

## Vector

@dataclass(init=False)
class FillProps(HasProps):

    fill_color: ColorSpec = ...
    fill_alpha: AlphaSpec = ...

@dataclass(init=False)
class HatchProps(HasProps):

    hatch_color: ColorSpec = ...
    hatch_alpha: AlphaSpec = ...
    hatch_scale: FloatSpec = ...
    hatch_pattern: HatchPatternSpec = ...
    hatch_weight: FloatSpec = ...
    hatch_extra: dict[str, Texture] = ...

@dataclass(init=False)
class ImageProps(HasProps):

    global_alpha: AlphaSpec = ...

@dataclass(init=False)
class LineProps(HasProps):

    line_color: ColorSpec = ...
    line_alpha: AlphaSpec = ...
    line_width: FloatSpec = ...
    line_join: LineJoinSpec = ...
    line_cap: LineCapSpec = ...
    line_dash: DashPatternSpec = ...
    line_dash_offset: IntSpec = ...

@dataclass(init=False)
class TextProps(HasProps):

    text_color: ColorSpec = ...
    text_outline_color: ColorSpec = ...
    text_alpha: AlphaSpec = ...
    text_font: StringSpec = ...
    text_font_size: FontSizeSpec = ...
    text_font_style: FontStyleSpec = ...
    text_align: TextAlignSpec = ...
    text_baseline: TextBaselineSpec = ...
    text_line_height: NumberSpec = ...

# background_

@dataclass(init=False)
class BackgroundFillProps(HasProps):

    background_fill_color: ColorSpec = ...
    background_fill_alpha: AlphaSpec = ...

@dataclass(init=False)
class BackgroundHatchProps(HasProps):

    background_hatch_color: ColorSpec = ...
    background_hatch_alpha: AlphaSpec = ...
    background_hatch_scale: FloatSpec = ...
    background_hatch_pattern: HatchPatternSpec = ...
    background_hatch_weight: FloatSpec = ...
    background_hatch_extra: dict[str, Texture] = ...

# border_

@dataclass(init=False)
class BorderLineProps(HasProps):

    border_line_color: ColorSpec = ...
    border_line_alpha: AlphaSpec = ...
    border_line_width: FloatSpec = ...
    border_line_join: LineJoinSpec = ...
    border_line_cap: LineCapSpec = ...
    border_line_dash: DashPatternSpec = ...
    border_line_dash_offset: IntSpec = ...

# body_

@dataclass(init=False)
class BodyLineProps(HasProps):

    body_line_color: ColorSpec = ...
    body_line_alpha: AlphaSpec = ...
    body_line_width: FloatSpec = ...
    body_line_join: LineJoinSpec = ...
    body_line_cap: LineCapSpec = ...
    body_line_dash: DashPatternSpec = ...
    body_line_dash_offset: IntSpec = ...

# Scalar

@dataclass(init=False)
class ScalarFillProps(HasProps):

    fill_color: Color | None = ...
    fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarHatchProps(HasProps):

    hatch_color: Color | None = ...
    hatch_alpha: Alpha = ...
    hatch_scale: Size = ...
    hatch_pattern: str | None = ...
    hatch_weight: Size = ...
    hatch_extra: dict[str, Texture] = ...

@dataclass(init=False)
class ScalarImageProps(HasProps):

    global_alpha: Alpha = ...

@dataclass(init=False)
class ScalarLineProps(HasProps):

    line_color: Color | None = ...
    line_alpha: Alpha = ...
    line_width: float = ...
    line_join: LineJoin = ...
    line_cap: LineCap = ...
    line_dash: DashPattern = ...
    line_dash_offset: int = ...

@dataclass(init=False)
class ScalarTextProps(HasProps):

    text_color: Color | None = ...
    text_outline_color: Color | None = ...
    text_alpha: Alpha = ...
    text_font: str = ...
    text_font_size: FontSize = ...
    text_font_style: FontStyle = ...
    text_align: TextAlign = ...
    text_baseline: TextBaseline = ...
    text_line_height: float = ...

# above_

@dataclass(init=False)
class ScalarAboveFillProps(HasProps):

    above_fill_color: Color | None = ...
    above_fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarAboveHatchProps(HasProps):

    above_hatch_color: Color | None = ...
    above_hatch_alpha: Alpha = ...
    above_hatch_scale: Size = ...
    above_hatch_pattern: str | None = ...
    above_hatch_weight: Size = ...
    above_hatch_extra: dict[str, Texture] = ...

# axis_label_

@dataclass(init=False)
class ScalarAxisLabelTextProps(HasProps):

    axis_label_text_color: Color | None = ...
    axis_label_text_outline_color: Color | None = ...
    axis_label_text_alpha: Alpha = ...
    axis_label_text_font: str = ...
    axis_label_text_font_size: FontSize = ...
    axis_label_text_font_style: FontStyle = ...
    axis_label_text_align: TextAlign = ...
    axis_label_text_baseline: TextBaseline = ...
    axis_label_text_line_height: float = ...

# axis_

@dataclass(init=False)
class ScalarAxisLineProps(HasProps):

    axis_line_color: Color | None = ...
    axis_line_alpha: Alpha = ...
    axis_line_width: float = ...
    axis_line_join: LineJoin = ...
    axis_line_cap: LineCap = ...
    axis_line_dash: DashPattern = ...
    axis_line_dash_offset: int = ...

# background_

@dataclass(init=False)
class ScalarBackgroundFillProps(HasProps):

    background_fill_color: Color | None = ...
    background_fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarBackgroundHatchProps(HasProps):

    background_hatch_color: Color | None = ...
    background_hatch_alpha: Alpha = ...
    background_hatch_scale: Size = ...
    background_hatch_pattern: str | None = ...
    background_hatch_weight: Size = ...
    background_hatch_extra: dict[str, Texture] = ...

# band_

@dataclass(init=False)
class ScalarBandFillProps(HasProps):

    band_fill_color: Color | None = ...
    band_fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarBandHatchProps(HasProps):

    band_hatch_color: Color | None = ...
    band_hatch_alpha: Alpha = ...
    band_hatch_scale: Size = ...
    band_hatch_pattern: str | None = ...
    band_hatch_weight: Size = ...
    band_hatch_extra: dict[str, Texture] = ...

# bar_

@dataclass(init=False)
class ScalarBarLineProps(HasProps):

    bar_line_color: Color | None = ...
    bar_line_alpha: Alpha = ...
    bar_line_width: float = ...
    bar_line_join: LineJoin = ...
    bar_line_cap: LineCap = ...
    bar_line_dash: DashPattern = ...
    bar_line_dash_offset: int = ...

# below_

@dataclass(init=False)
class ScalarBelowFillProps(HasProps):

    below_fill_color: Color | None = ...
    below_fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarBelowHatchProps(HasProps):

    below_hatch_color: Color | None = ...
    below_hatch_alpha: Alpha = ...
    below_hatch_scale: Size = ...
    below_hatch_pattern: str | None = ...
    below_hatch_weight: Size = ...
    below_hatch_extra: dict[str, Texture] = ...

# border_

@dataclass(init=False)
class ScalarBorderLineProps(HasProps):

    border_line_color: Color | None = ...
    border_line_alpha: Alpha = ...
    border_line_width: float = ...
    border_line_join: LineJoin = ...
    border_line_cap: LineCap = ...
    border_line_dash: DashPattern = ...
    border_line_dash_offset: int = ...

@dataclass(init=False)
class ScalarBorderFillProps(HasProps):

    border_fill_color: Color | None = ...
    border_fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarBorderHatchProps(HasProps):

    border_hatch_color: Color | None = ...
    border_hatch_alpha: Alpha = ...
    border_hatch_scale: Size = ...
    border_hatch_pattern: str | None = ...
    border_hatch_weight: Size = ...
    border_hatch_extra: dict[str, Texture] = ...

# grid_

@dataclass(init=False)
class ScalarGridLineProps(HasProps):

    grid_line_color: Color | None = ...
    grid_line_alpha: Alpha = ...
    grid_line_width: float = ...
    grid_line_join: LineJoin = ...
    grid_line_cap: LineCap = ...
    grid_line_dash: DashPattern = ...
    grid_line_dash_offset: int = ...


# group_

@dataclass(init=False)
class ScalarGroupTextProps(HasProps):

    group_text_color: Color | None = ...
    group_text_outline_color: Color | None = ...
    group_text_alpha: Alpha = ...
    group_text_font: str = ...
    group_text_font_size: FontSize = ...
    group_text_font_style: FontStyle = ...
    group_text_align: TextAlign = ...
    group_text_baseline: TextBaseline = ...
    group_text_line_height: float = ...

# hover_

@dataclass(init=False)
class ScalarHoverFillProps(HasProps):

    hover_fill_color: Color | None = ...
    hover_fill_alpha: Alpha = ...

@dataclass(init=False)
class ScalarHoverHatchProps(HasProps):

    hover_hatch_color: Color | None = ...
    hover_hatch_alpha: Alpha = ...
    hover_hatch_scale: Size = ...
    hover_hatch_pattern: str | None = ...
    hover_hatch_weight: Size = ...
    hover_hatch_extra: dict[str, Texture] = ...

@dataclass(init=False)
class ScalarHoverLineProps(HasProps):

    hover_line_color: Color | None = ...
    hover_line_alpha: Alpha = ...
    hover_line_width: float = ...
    hover_line_join: LineJoin = ...
    hover_line_cap: LineCap = ...
    hover_line_dash: DashPattern = ...
    hover_line_dash_offset: int = ...

# inactive_

@dataclass(init=False)
class ScalarInactiveFillProps(HasProps):

    inactive_fill_color: Color | None = ...
    inactive_fill_alpha: Alpha = ...

# item_background_

@dataclass(init=False)
class ScalarItemBackgroundFillProps(HasProps):

    item_background_fill_color: Color | None = ...
    item_background_fill_alpha: Alpha = ...

# label_

@dataclass(init=False)
class ScalarLabelTextProps(HasProps):

    label_text_color: Color | None = ...
    label_text_outline_color: Color | None = ...
    label_text_alpha: Alpha = ...
    label_text_font: str = ...
    label_text_font_size: FontSize = ...
    label_text_font_style: FontStyle = ...
    label_text_align: TextAlign = ...
    label_text_baseline: TextBaseline = ...
    label_text_line_height: float = ...

# major_label_

@dataclass(init=False)
class ScalarMajorLabelTextProps(HasProps):

    major_label_text_color: Color | None = ...
    major_label_text_outline_color: Color | None = ...
    major_label_text_alpha: Alpha = ...
    major_label_text_font: str = ...
    major_label_text_font_size: FontSize = ...
    major_label_text_font_style: FontStyle = ...
    major_label_text_align: TextAlign = ...
    major_label_text_baseline: TextBaseline = ...
    major_label_text_line_height: float = ...

# major_tick_

@dataclass(init=False)
class ScalarMajorTickLineProps(HasProps):

    major_tick_line_color: Color | None = ...
    major_tick_line_alpha: Alpha = ...
    major_tick_line_width: float = ...
    major_tick_line_join: LineJoin = ...
    major_tick_line_cap: LineCap = ...
    major_tick_line_dash: DashPattern = ...
    major_tick_line_dash_offset: int = ...

# minor_grid_

@dataclass(init=False)
class ScalarMinorGridLineProps(HasProps):

    minor_grid_line_color: Color | None = ...
    minor_grid_line_alpha: Alpha = ...
    minor_grid_line_width: float = ...
    minor_grid_line_join: LineJoin = ...
    minor_grid_line_cap: LineCap = ...
    minor_grid_line_dash: DashPattern = ...
    minor_grid_line_dash_offset: int = ...


# minor_tick_

@dataclass(init=False)
class ScalarMinorTickLineProps(HasProps):

    minor_tick_line_color: Color | None = ...
    minor_tick_line_alpha: Alpha = ...
    minor_tick_line_width: float = ...
    minor_tick_line_join: LineJoin = ...
    minor_tick_line_cap: LineCap = ...
    minor_tick_line_dash: DashPattern = ...
    minor_tick_line_dash_offset: int = ...

# outline_

@dataclass(init=False)
class ScalarOutlineLineProps(HasProps):

    outline_line_color: Color | None = ...
    outline_line_alpha: Alpha = ...
    outline_line_width: float = ...
    outline_line_join: LineJoin = ...
    outline_line_cap: LineCap = ...
    outline_line_dash: DashPattern = ...
    outline_line_dash_offset: int = ...

# separator_

@dataclass(init=False)
class ScalarSeparatorLineProps(HasProps):

    separator_line_color: Color | None = ...
    separator_line_alpha: Alpha = ...
    separator_line_width: float = ...
    separator_line_join: LineJoin = ...
    separator_line_cap: LineCap = ...
    separator_line_dash: DashPattern = ...
    separator_line_dash_offset: int = ...

# subgroup_

@dataclass(init=False)
class ScalarSubgroupTextProps(HasProps):

    subgroup_text_color: Color | None = ...
    subgroup_text_outline_color: Color | None = ...
    subgroup_text_alpha: Alpha = ...
    subgroup_text_font: str = ...
    subgroup_text_font_size: FontSize = ...
    subgroup_text_font_style: FontStyle = ...
    subgroup_text_align: TextAlign = ...
    subgroup_text_baseline: TextBaseline = ...
    subgroup_text_line_height: float = ...

# title_

@dataclass(init=False)
class ScalarTitleTextProps(HasProps):

    title_text_color: Color | None = ...
    title_text_outline_color: Color | None = ...
    title_text_alpha: Alpha = ...
    title_text_font: str = ...
    title_text_font_size: FontSize = ...
    title_text_font_style: FontStyle = ...
    title_text_align: TextAlign = ...
    title_text_baseline: TextBaseline = ...
    title_text_line_height: float = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
""" Reusable common property type aliases.

"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Bokeh imports
from . import enums
from .property.auto import Auto
from .property.container import Dict, List, Tuple
from .property.either import Either
from .property.enum import Enum
from .property.numeric import Int, NonNegative, Percent
from .property.string import Regex, String
from .property.struct import Optional, Struct
from .property.visual import Image

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "Anchor",
    "AutoAnchor",
    "BorderRadius",
    "CSSClass",
    "CSSVariable",
    "DataImage",
    "GridSpacing",
    "IconLike",
    "Padding",
    "Pixels",
    "TextAnchor",
    "TracksSizing",
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

CSSVariable = Regex(r"^--")

CSSClass = Regex(r"^\.")

DataImage = Regex(r"^\data:image")

IconLike = Either(Image, Enum(enums.ToolIcon), CSSVariable, CSSClass, DataImage)

Pixels = NonNegative(Int)

HAnchor = Either(Enum(enums.Align), Enum(enums.HAlign), Percent)
VAnchor = Either(Enum(enums.Align), Enum(enums.VAlign), Percent)

Anchor = (
    Either(
        Enum(enums.Anchor),
        Tuple(HAnchor, VAnchor),
    )
)

AutoAnchor = (
    Either(
        Auto,
        Enum(enums.Anchor),
        Tuple(Either(Auto, HAnchor), Either(Auto, VAnchor)),
    )
)

TextAnchor = Either(Anchor, Auto)

BorderRadius = (
    Either(
        Pixels,
        Tuple(Pixels, Pixels, Pixels, Pixels),
        Struct(
            top_left=Optional(Pixels),
            top_right=Optional(Pixels),
            bottom_right=Optional(Pixels),
            bottom_left=Optional(Pixels),
        ),
    )
)

Padding = (
    Either(
        Pixels,
        Tuple(Pixels, Pixels),
        Struct(
            x=Optional(Pixels),
            y=Optional(Pixels),
        ),
        Tuple(Pixels, Pixels, Pixels, Pixels),
        Struct(
            left=Optional(Pixels),
            right=Optional(Pixels),
            top=Optional(Pixels),
            bottom=Optional(Pixels),
        ),
    )
)

GridSpacing = Either(Pixels, Tuple(Pixels, Pixels))

TrackAlign = Enum("start", "center", "end", "auto")

# CSS length, percentage, flex, max-content, min-content, auto, etc.
# See https://developer.mozilla.org/en-US/docs/Web/CSS/grid-template-columns.
TrackSize = String()

TrackSizing = Either(TrackSize, Struct(size=Optional(TrackSize), align=Optional(TrackAlign)))

TracksSizing = Either(TrackSizing, List(TrackSizing), Dict(Int, TrackSizing))

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from typing import TypedDict, Unpack

# Bokeh imports
from .._specs import (
    AlphaArg,
    AngleArg,
    ColorArg,
    DashPatternArg,
    DistanceArg,
    FloatArg,
    FontSizeArg,
    FontStyleArg,
    HatchPatternArg,
    Image2dArg,
    Image3dArg,
    IntArg,
    LineCapArg,
    LineJoinArg,
    MarkerArg,
    NonNegative,
    NullDistanceArg,
    Number1dArg,
    Number3dArg,
    NumberArg,
    OutlineShapeNameArg,
    SizeArg,
    StringArg,
    TextAlignArg,
    TextAnchorArg,
    TextBaselineArg,
)
from .._types import Color
from ..core.enums import (
    AnchorType as Anchor,
    AngleUnitsType as AngleUnits,
    DirectionType as Direction,
    PaletteType as Palette,
    RadiusDimensionType as RadiusDimension,
    RenderLevelType as RendererLevel,
    SpatialUnitsType as SpatialUnits,
    StepModeType as StepMode,
    TeXDisplayType as TeXDisplay,
)
from ..core.property_aliases import (
    BorderRadiusType as BorderRadius,
    PaddingType as Padding,
)
from ..models import glyphs
from ..models.annotations import Legend
from ..models.callbacks import CustomJS
from ..models.coordinates import CoordinateMapping
from ..models.plots import Plot
from ..models.renderers import GlyphRenderer
from ..models.sources import CDSView, ColumnarDataSource, DataDictLike
from ..models.textures import Texture

class AuxVisuals(TypedDict, total=False):
    color: ColorArg
    alpha: AlphaArg

    selection_color: ColorArg
    selection_alpha: AlphaArg

    nonselection_color: ColorArg
    nonselection_alpha: AlphaArg

    hover_color: ColorArg
    hover_alpha: AlphaArg

    muted_color: ColorArg
    muted_alpha: AlphaArg

class AuxHatchVisuals(TypedDict, total=False):
    selection_hatch_color: ColorArg
    selection_hatch_alpha: AlphaArg

    nonselection_hatch_color: ColorArg
    nonselection_hatch_alpha: AlphaArg

    hover_hatch_color: ColorArg
    hover_hatch_alpha: AlphaArg

    muted_hatch_color: ColorArg
    muted_hatch_alpha: AlphaArg

class AuxFillVisuals(TypedDict, total=False):
    selection_fill_color: ColorArg
    selection_fill_alpha: AlphaArg

    nonselection_fill_color: ColorArg
    nonselection_fill_alpha: AlphaArg

    hover_fill_color: ColorArg
    hover_fill_alpha: AlphaArg

    muted_fill_color: ColorArg
    muted_fill_alpha: AlphaArg

class AuxImageVisuals(TypedDict, total=False):
    pass

class AuxLineVisuals(TypedDict, total=False):
    selection_line_color: ColorArg
    selection_line_alpha: AlphaArg

    nonselection_line_color: ColorArg
    nonselection_line_alpha: AlphaArg

    hover_line_color: ColorArg
    hover_line_alpha: AlphaArg

    muted_line_color: ColorArg
    muted_line_alpha: AlphaArg

class AuxTextVisuals(TypedDict, total=False):
    selection_text_color: ColorArg
    selection_text_alpha: AlphaArg

    nonselection_text_color: ColorArg
    nonselection_text_alpha: AlphaArg

    hover_text_color: ColorArg
    hover_text_alpha: AlphaArg

    muted_text_color: ColorArg
    muted_text_alpha: AlphaArg

class FillVisuals(AuxFillVisuals, total=False):
    fill_color: ColorArg
    fill_alpha: AlphaArg

class HatchVisuals(AuxHatchVisuals, total=False):
    hatch_color: ColorArg
    hatch_alpha: AlphaArg
    hatch_scale: FloatArg
    hatch_pattern: HatchPatternArg
    hatch_weight: FloatArg
    hatch_extra: dict[str, Texture]

class ImageVisuals(AuxImageVisuals, total=False):
    global_alpha: AlphaArg

class LineVisuals(AuxLineVisuals, total=False):
    line_color: ColorArg
    line_alpha: AlphaArg
    line_width: FloatArg
    line_join: LineJoinArg
    line_cap: LineCapArg
    line_dash: DashPatternArg
    line_dash_offset: IntArg

class TextVisuals(AuxTextVisuals, total=False):
    text_color: ColorArg
    text_outline_color: ColorArg
    text_alpha: AlphaArg
    text_font: StringArg
    text_font_size: FontSizeArg
    text_font_style: FontStyleArg
    text_align: TextAlignArg
    text_baseline: TextBaselineArg
    text_line_height: NumberArg

class AuxGlyphArgs(TypedDict, total=False):
    # Model
    name: str | None

    # Renderer
    coordinates: CoordinateMapping | None
    x_range_name: str
    y_range_name: str
    level: RendererLevel
    visible: bool

    # GlyphRenderer
    source: ColumnarDataSource | DataDictLike
    view: CDSView
    muted: bool

    legend: Legend
    legend_label: str
    legend_field: str
    legend_group: str

class GlyphArgs(AuxGlyphArgs, AuxVisuals, total=False):
    pass

class AnnularWedgeArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # inner_radius: DistanceArg
    inner_radius_units: SpatialUnits
    # outer_radius: DistanceArg
    outer_radius_units: SpatialUnits
    # start_angle: AngleArg
    start_angle_units: AngleUnits
    # end_angle: AngleArg
    end_angle_units: AngleUnits
    # direction: Direction

class AnnulusArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # inner_radius: DistanceArg
    inner_radius_units: SpatialUnits
    # outer_radius: DistanceArg
    outer_radius_units: SpatialUnits

class ArcArgs(GlyphArgs, LineVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # radius: DistanceArg
    radius_units: SpatialUnits
    # start_angle: AngleArg
    start_angle_units: AngleUnits
    # end_angle: AngleArg
    end_angle_units: AngleUnits
    # direction: Direction

class BezierArgs(GlyphArgs, LineVisuals, total=False):
    # x0: NumberArg
    # y0: NumberArg
    # x1: NumberArg
    # y1: NumberArg
    # cx0: NumberArg
    # cy0: NumberArg
    # cx1: NumberArg
    # cy1: NumberArg
    pass

class BlockArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # width: DistanceArg
    width_units: SpatialUnits
    # height: DistanceArg
    height_units: SpatialUnits

class CircleArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # radius: DistanceArg
    radius_units: SpatialUnits
    radius_dimension: RadiusDimension
    hit_dilation: NonNegative[float]

class EllipseArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # width: DistanceArg
    width_units: SpatialUnits
    # height: DistanceArg
    height_units: SpatialUnits
    # angle: AngleArg
    angel_units: AngleUnits

class HAreaArgs(GlyphArgs, FillVisuals, HatchVisuals, total=False):
    # x1: NumberArg
    # x2: NumberArg
    # y: NumberArg
    pass

class HAreaStepArgs(GlyphArgs, FillVisuals, HatchVisuals, total=False):
    # x1: NumberArg
    # x2: NumberArg
    # y: NumberArg
    step_mode: StepMode

class HBarArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # y: NumberArg
    # height: DistanceArg
    height_units: SpatialUnits
    # left: NumberArg
    # right: NumberArg

class HSpanArgs(GlyphArgs, LineVisuals, total=False):
    # y: NumberArg
    pass

class HStripArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # y0: NumberArg
    # y1: NumberArg
    pass

class HexTileArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # r: NumberArg
    # q: NumberArg
    size: float
    aspect_scale: float
    scale: NumberArg
    orientation: str

class ImageArgs(GlyphArgs, total=False):
    # image: Image2dArg
    # x: NumberArg
    # y: NumberArg
    # dw: DistanceArg
    # dh: DistanceArg
    # dilate: bool
    palette: Palette | list[Color]

class ImageRGBAArgs(GlyphArgs, total=False):
    # image: Image2dArg
    # x: NumberArg
    # y: NumberArg
    # dw: DistanceArg
    # dh: DistanceArg
    # dilate: bool
    pass

class ImageStackArgs(GlyphArgs, total=False):
    # image: Image3dArg
    # x: NumberArg
    # y: NumberArg
    # dw: DistanceArg
    # dh: DistanceArg
    # dilate: bool
    pass

class ImageURLArgs(GlyphArgs, total=False):
    # url: StringArg
    # x: NumberArg
    # y: NumberArg
    # w: NullDistanceArg
    w_units: SpatialUnits
    # h: NullDistanceArg
    h_units: SpatialUnits
    # angle: AngleArg
    angle_units: AngleUnits
    global_alpha: NumberArg
    # dilate: bool
    anchor: Anchor
    retry_attempts: int
    retry_timeout: int

class LineArgs(GlyphArgs, LineVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    pass

class MarkerArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # size: SizeArg
    # angle: AngleArg
    hit_dilation: NonNegative[float]

class MathMLArgs(GlyphArgs, TextVisuals, total=False):
    pass

class MultiLineArgs(GlyphArgs, LineVisuals, total=False):
    # xs: Number1dArg
    # ys: Number1dArg
    pass

class MultiPolygonsArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # xs: Number3dArg
    # ys: Number3dArg
    pass

class NgonArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # radius: DistanceArg
    radius_units: SpatialUnits
    # angle: AngleArg
    angle_units: AngleUnits
    n: NumberArg
    radius_dimension: RadiusDimension

class PatchArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    pass

class PatchesArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # xs: Number1dArg
    # ys: Number1dArg
    pass

class QuadArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # left: NumberArg
    # right: NumberArg
    # bottom: NumberArg
    # top: NumberArg
    pass

class QuadraticArgs(GlyphArgs, LineVisuals, total=False):
    # x0: NumberArg
    # y0: NumberArg
    # x1: NumberArg
    # y1: NumberArg
    # cx: NumberArg
    # cy: NumberArg
    pass

class RayArgs(GlyphArgs, LineVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # length: DistanceArg
    length_units: SpatialUnits
    # angle: AngleArg
    angle_units: AngleUnits

class RectArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # width: DistanceArg
    width_units: SpatialUnits
    # height: DistanceArg
    height_units: SpatialUnits
    # angle: AngleArg
    angle_units: AngleUnits
    # dilate: bool
    border_radius: BorderRadius

class ScatterArgs(MarkerArgs, total=False):
    # marker: MarkerArg
    defs: dict[str, CustomJS]

class SegmentArgs(GlyphArgs, LineVisuals, total=False):
    # x0: NumberArg
    # y0: NumberArg
    # x1: NumberArg
    # y1: NumberArg
    pass

class StepArgs(GlyphArgs, LineVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    mode: StepMode

class TeXArgs(GlyphArgs, TextVisuals, total=False):
    macros: dict[str, str | tuple[str, int]]
    display: TeXDisplay

class TextArgs(GlyphArgs, TextVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # text: StringArg
    # angle: AngleArg
    angle_units: AngleUnits
    # x_offset: FloatArg
    # y_offset: FloatArg
    anchor: TextAnchorArg
    padding: Padding
    border_radius: BorderRadius
    outline_shape: OutlineShapeNameArg

class VAreaArgs(GlyphArgs, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y1: NumberArg
    # y2: NumberArg
    pass

class VAreaStepArgs(GlyphArgs, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y1: NumberArg
    # y2: NumberArg
    step_mode: StepMode

class VBarArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # width: DistanceArg
    width_units: SpatialUnits
    # bottom: NumberArg
    # top: NumberArg

class VSpanArgs(GlyphArgs, LineVisuals, total=False):
    # x: NumberArg
    pass

class VStripArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x0: NumberArg
    # x1: NumberArg
    pass

class WedgeArgs(GlyphArgs, LineVisuals, FillVisuals, HatchVisuals, total=False):
    # x: NumberArg
    # y: NumberArg
    # radius: DistanceArg
    radius_units: SpatialUnits
    # start_angle: AngleArg
    start_angle_units: AngleUnits
    # end_angle: AngleArg
    end_angle_units: AngleUnits
    # direction: Direction

class GlyphAPI:

    @property
    def plot(self) -> Plot | None: ...

    @property
    def coordinates(self) -> CoordinateMapping | None: ...

    def __init__(self, parent: Plot | None = None, coordinates: CoordinateMapping | None = None) -> None: ...

    def annular_wedge(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        inner_radius: DistanceArg = ...,
        outer_radius: DistanceArg = ...,
        start_angle: AngleArg = ...,
        end_angle: AngleArg = ...,
        direction: Direction = ...,
        **kwargs: Unpack[AnnularWedgeArgs],
    ) -> GlyphRenderer[glyphs.AnnularWedge]: ...

    def annulus(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        inner_radius: DistanceArg = ...,
        outer_radius: DistanceArg = ...,
        **kwargs: Unpack[AnnulusArgs],
    ) -> GlyphRenderer[glyphs.Annulus]: ...

    def arc(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        radius: DistanceArg = ...,
        start_angle: AngleArg = ...,
        end_angle: AngleArg = ...,
        direction: Direction = ...,
        **kwargs: Unpack[ArcArgs],
    ) -> GlyphRenderer[glyphs.Arc]: ...

    def bezier(self,
        x0: NumberArg = ...,
        y0: NumberArg = ...,
        x1: NumberArg = ...,
        y1: NumberArg = ...,
        cx0: NumberArg = ...,
        cy0: NumberArg = ...,
        cx1: NumberArg = ...,
        cy1: NumberArg = ...,
        **kwargs: Unpack[BezierArgs],
    ) -> GlyphRenderer[glyphs.Bezier]: ...

    def circle(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        radius: DistanceArg = ...,
        **kwargs: Unpack[CircleArgs],
    ) -> GlyphRenderer[glyphs.Circle]: ...

    def block(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        width: DistanceArg = ...,
        height: DistanceArg = ...,
        **kwargs: Unpack[BlockArgs],
    ) -> GlyphRenderer[glyphs.Block]: ...

    def harea(self,
        x1: NumberArg = ...,
        x2: NumberArg = ...,
        y: NumberArg = ...,
        **kwargs: Unpack[HAreaArgs],
    ) -> GlyphRenderer[glyphs.HArea]: ...

    def harea_step(self,
        x1: NumberArg = ...,
        x2: NumberArg = ...,
        y: NumberArg = ...,
        **kwargs: Unpack[HAreaStepArgs],
    ) -> GlyphRenderer[glyphs.HAreaStep]: ...

    def hbar(self,
        y: NumberArg = ...,
        height: DistanceArg = ...,
        right: NumberArg = ...,
        left: NumberArg = ...,
        **kwargs: Unpack[HBarArgs],
    ) -> GlyphRenderer[glyphs.HBar]: ...

    def hspan(self,
        y: NumberArg = ...,
        **kwargs: Unpack[HSpanArgs],
    ) -> GlyphRenderer[glyphs.HSpan]: ...

    def hstrip(self,
        y0: NumberArg = ...,
        y1: NumberArg = ...,
        **kwargs: Unpack[HStripArgs],
    ) -> GlyphRenderer[glyphs.HStrip]: ...

    def ellipse(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        width: DistanceArg = ...,
        height: DistanceArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[EllipseArgs],
    ) -> GlyphRenderer[glyphs.Ellipse]: ...

    def hex_tile(self,
        q: NumberArg = ...,
        r: NumberArg = ...,
        **kwargs: Unpack[HexTileArgs],
    ) -> GlyphRenderer[glyphs.HexTile]: ...

    def image(self,
        image: Image2dArg = ...,
        x: NumberArg = ...,
        y: NumberArg = ...,
        dw: DistanceArg = ...,
        dh: DistanceArg = ...,
        dilate: bool = ...,
        **kwargs: Unpack[ImageArgs],
    ) -> GlyphRenderer[glyphs.Image]: ...

    def image_rgba(self,
        image: Image2dArg = ...,
        x: NumberArg = ...,
        y: NumberArg = ...,
        dw: DistanceArg = ...,
        dh: DistanceArg = ...,
        dilate: bool = ...,
        **kwargs: Unpack[ImageRGBAArgs],
    ) -> GlyphRenderer[glyphs.ImageRGBA]: ...

    def image_stack(self,
        image: Image3dArg = ...,
        x: NumberArg = ...,
        y: NumberArg = ...,
        dw: DistanceArg = ...,
        dh: DistanceArg = ...,
        dilate: bool = ...,
        **kwargs: Unpack[ImageStackArgs],
    ) -> GlyphRenderer[glyphs.ImageStack]: ...

    def image_url(self,
        url: StringArg = ...,
        x: NumberArg = ...,
        y: NumberArg = ...,
        w: NullDistanceArg = ...,
        h: NullDistanceArg = ...,
        angle: AngleArg = ...,
        dilate: bool = ...,
        **kwargs: Unpack[ImageURLArgs],
    ) -> GlyphRenderer[glyphs.ImageURL]: ...

    def line(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        **kwargs: Unpack[LineArgs],
    ) -> GlyphRenderer[glyphs.Line]: ...

    def mathml(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        text: StringArg = ...,
        angle: AngleArg = ...,
        x_offset: FloatArg = ...,
        y_offset: FloatArg = ...,
        **kwargs: Unpack[MathMLArgs],
    ) -> GlyphRenderer[glyphs.MathMLGlyph]: ...

    def multi_line(self,
        xs: Number1dArg = ...,
        ys: Number1dArg = ...,
        **kwargs: Unpack[MultiLineArgs],
    ) -> GlyphRenderer[glyphs.MultiLine]: ...

    def multi_polygons(self,
        xs: Number3dArg = ...,
        ys: Number3dArg = ...,
        **kwargs: Unpack[MultiPolygonsArgs],
    ) -> GlyphRenderer[glyphs.MultiPolygons]: ...

    def ngon(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        radius: DistanceArg = ...,
        **kwargs: Unpack[NgonArgs],
    ) -> GlyphRenderer[glyphs.Ngon]: ...

    def patch(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        **kwargs: Unpack[PatchArgs],
    ) -> GlyphRenderer[glyphs.Patch]: ...

    def patches(self,
        xs: Number1dArg = ...,
        ys: Number1dArg = ...,
        **kwargs: Unpack[PatchesArgs],
    ) -> GlyphRenderer[glyphs.Patches]: ...

    def quad(self,
        left: NumberArg = ...,
        right: NumberArg = ...,
        top: NumberArg = ...,
        bottom: NumberArg = ...,
        **kwargs: Unpack[QuadArgs],
    ) -> GlyphRenderer[glyphs.Quad]: ...

    def quadratic(self,
        x0: NumberArg = ...,
        y0: NumberArg = ...,
        x1: NumberArg = ...,
        y1: NumberArg = ...,
        cx: NumberArg = ...,
        cy: NumberArg = ...,
        **kwargs: Unpack[QuadraticArgs],
    ) -> GlyphRenderer[glyphs.Quadratic]: ...

    def ray(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        length: DistanceArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[RayArgs],
    ) -> GlyphRenderer[glyphs.Ray]: ...

    def rect(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        width: DistanceArg = ...,
        height: DistanceArg = ...,
        angle: AngleArg = ...,
        dilate: bool = ...,
        **kwargs: Unpack[RectArgs],
    ) -> GlyphRenderer[glyphs.Rect]: ...

    def step(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        **kwargs: Unpack[StepArgs],
    ) -> GlyphRenderer[glyphs.Step]: ...

    def segment(self,
        x0: NumberArg = ...,
        y0: NumberArg = ...,
        x1: NumberArg = ...,
        y1: NumberArg = ...,
        **kwargs: Unpack[SegmentArgs],
    ) -> GlyphRenderer[glyphs.Segment]: ...

    def tex(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        text: StringArg = ...,
        angle: AngleArg = ...,
        x_offset: FloatArg = ...,
        y_offset: FloatArg = ...,
        **kwargs: Unpack[TeXArgs],
    ) -> GlyphRenderer[glyphs.TeXGlyph]: ...

    def text(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        text: StringArg = ...,
        angle: AngleArg = ...,
        x_offset: FloatArg = ...,
        y_offset: FloatArg = ...,
        **kwargs: Unpack[TextArgs],
    ) -> GlyphRenderer[glyphs.Text]: ...

    def varea(self,
        x: NumberArg = ...,
        y1: NumberArg = ...,
        y2: NumberArg = ...,
        **kwargs: Unpack[VAreaArgs],
    ) -> GlyphRenderer[glyphs.VArea]: ...

    def varea_step(self,
        x: NumberArg = ...,
        y1: NumberArg = ...,
        y2: NumberArg = ...,
        **kwargs: Unpack[VAreaStepArgs],
    ) -> GlyphRenderer[glyphs.VAreaStep]: ...

    def vbar(self,
        x: NumberArg = ...,
        width: DistanceArg = ...,
        top: NumberArg = ...,
        bottom: NumberArg = ...,
        **kwargs: Unpack[VBarArgs],
    ) -> GlyphRenderer[glyphs.VBar]: ...

    def vspan(self,
        x: NumberArg = ...,
        **kwargs: Unpack[VSpanArgs],
    ) -> GlyphRenderer[glyphs.VSpan]: ...

    def vstrip(self,
        x0: NumberArg = ...,
        x1: NumberArg = ...,
        **kwargs: Unpack[VStripArgs],
    ) -> GlyphRenderer[glyphs.VStrip]: ...

    def wedge(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        radius: DistanceArg = ...,
        start_angle: AngleArg = ...,
        end_angle: AngleArg = ...,
        direction: Direction = ...,
        **kwargs: Unpack[WedgeArgs],
    ) -> GlyphRenderer[glyphs.Wedge]: ...

    def scatter(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        marker: MarkerArg = ...,
        **kwargs: Unpack[ScatterArgs],
    ) -> GlyphRenderer[glyphs.Scatter]: ...

    # Markers

    def asterisk(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def circle_cross(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def circle_dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def circle_x(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def circle_y(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def cross(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def dash(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def diamond(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def diamond_cross(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def diamond_dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def hex(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def hex_dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def inverted_triangle(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def plus(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def square(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def square_cross(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def square_dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def square_pin(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def square_x(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def star(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def star_dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def triangle(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def triangle_dot(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def triangle_pin(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def x(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...

    def y(self,
        x: NumberArg = ...,
        y: NumberArg = ...,
        size: SizeArg = ...,
        angle: AngleArg = ...,
        **kwargs: Unpack[MarkerArgs],
    ) -> GlyphRenderer[glyphs.Marker]: ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal

# Bokeh imports
from ...core.enums import OrientationType as Orientation
from ...core.has_props import abstract
from .widget import Widget

@abstract
@dataclass(init=False)
class Indicator(Widget):
    pass

@dataclass()
class Progress(Indicator):

    mode: Literal["determinate", "indeterminate"] = ...

    value: int = ...

    min: int = ...

    max: int = ...

    reversed: bool = ...

    orientation: Orientation = ...

    label: str | None = ...

    label_location: Literal["none", "inline"] = ...

    description: str | None = ...
